!==================================================================================================================================
! LVDSMC_VHS_2D.f90
!==================================================================================================================================

!   Low-variance direct simulation Monte Carlo (LVDSMC) solver.
!   written by: Gregg A. Radtke

!----------------------------------------------------------------------------------------------------------------------------------
! DISCLAIMER: This computational code is provided as is without any guarantees or warranty.  Although the author has attempted to
!   find and correct any bugs, the author is not responsible for any damage or losses of any kind caused by the use or misuse of
!   the code.  The author is under no obligation to provide support, service, corrections, or upgrades.
!----------------------------------------------------------------------------------------------------------------------------------

!----------------------------------------------------------------------------------------------------------------------------------
! REQUIRES:
!----------------------------------------------------------------------------------------------------------------------------------

!   (1) modern FORTRAN compiler
!   (2) input file ("input.inp"), in the format described below
!   (3) ratio-of-uniforms bounding factors file ("bounds.inp"), in the format described below

!   The program can be compiled with any modern FORTRAN compiler.  A free FORTRAN compiler (G95) is available from 
!   http://www.g95.org.  This code has been successfully compiled and run on G95 version 0.92.


!----------------------------------------------------------------------------------------------------------------------------------
! FEATURES:
!----------------------------------------------------------------------------------------------------------------------------------

!   Linearized variable hard sphere (VHS) collision model
!   Two-dimensional, rectangular geometry
!   Boundary conditions: Maxwell's accommodation model
!   Streamwise pressure and temperature gradients
!   Dimensionless variables


!----------------------------------------------------------------------------------------------------------------------------------
! VERSION HISTORY:
!----------------------------------------------------------------------------------------------------------------------------------

!   1.0      Jan 26, 2011


!----------------------------------------------------------------------------------------------------------------------------------
! DIMENSIONLESS VARIABLES:
!----------------------------------------------------------------------------------------------------------------------------------

!   dimensional parameters
!   ----------------------
!   f0       reference equilibrium distribution [kg s^3/m^6]
!   rho0     reference density [kg/m^3]
!   T0       reference temperature [K]
!   c0       most probable molecular speed at reference conditions [m/s]
!   l0       mean free path at reference condtions [m]
!   d0       effective diameter at reference conditions [m]
!   m        molecular mass [kg]
!   L0       system length scale [m]
!   R        gas constant ( = kB/m, where kB is the Boltzmann constant) [J/(kg K)]
!   c        molecular velocity (vector) [m/s]
!   x        spatial position (0 <= x <= L0)
!   rho      mass density [kg/m^3]
!   u        velocity [m/s]
!   P        pressure tensor [N/m]
!   T        temperature [K]
!   q        heat flux [W/m^2]

!   reference equilibrium distribution (also the initial condition)
!   ---------------------------------------------------------------
!   f0 = rho0 / ( pi^(3/2) c0^3) exp ( - || c ||^2 / c0^2 )
!     where c0 = sqrt ( 2 R T0 )

!   Knudsen number:
!   ---------------
!   Kn = l0 / L0
!     where 1 / l0 = sqrt(2) pi ( rho0 / m ) d0^2

!   dimensionless spatial position, time, particle velocity
!   -------------------------------------------------------
!   x* = x / L0
!   t* = c0 t / L0
!   c* = c / c0

!   dimensionless hydrodynamic variables
!   ------------------------------------
!   rho* = ( rho / rho0 - 1 ) / epsilon
!   u*   = ( u   / c0 ) / epsilon
!   P*   = [ P   / ( rho0 R T0 )   - I ] / epsilon     (where I is the 3x3 identity tensor)
!   T*   = ( T   / T0   - 1 ) / epsilon
!   q*   = [ q   / ( c0 rho0 R T0 ) ] / epsilon


!----------------------------------------------------------------------------------------------------------------------------------
! INPUT FILE:     ( file name = "input.inp" )
!----------------------------------------------------------------------------------------------------------------------------------

!   input file format
!   -----------------
!   Nt, dt, Nout, Nss
!   Nx, Ny, Lx, Ly, Xi, Nens, disp
!   Kn, omega, epsilon, kappa_P, kappa_T
!   type_b1, alpha_b1, ux*_b1, uy*_b1, uz*_b1, T*_b1
!   type_b2, alpha_b2, ux*_b2, uy*_b2, uz*_b2, T*_b2
!      .         .        .       .       .
!      .         .        .       .       .
!      .         .        .       .       .
!   type_bM, alpha_bM, ux*_bM, uy*_bM, uz*_bM, T*_bM

!   geometry
!   --------

!     There are M = 2Nx + 2Ny boundary conditions numbered in a counterclockwise fashion as shown below.

!             2Nx+Ny            ...   Nx+Ny+2 Nx+Ny+1
!            +-------+-------+  ...  +-------+-------+
!            |       |       |       |       |       |
!   2Nx+Ny+1 |       |       |       |       |       | Nx+Ny
!            |       |       |       |       |       |
!            +-------+-------+  ...  +-------+-------+
!            |       |       |       |       |       |
!   2Nx+Ny+2 |       |       |       |       |       |
!            |       |       |       |       |       |
!            +-------+-------+  ...  +-------+-------+
!      .     .       .       .       .       .       .  .
!      .     .       .       .       .       .       .  .
!      .     .       .       .       .       .       .  .
!            +-------+-------+  ...  +-------+-------+
!            |       |       |       |       |       |
!            |       |       |       |       |       | Nx+2
!            |       |       |       |       |       |
!            +-------+-------+  ...  +-------+-------+
!            |       |       |       |       |       |
!    2Nx+2Ny |       |       |       |       |       | Nx+1
!            |       |       |       |       |       |
!            +-------+-------+  ...  +-------+-------+
!                1       2      ...             Nx

!     The boundary velocities in the input files are referenced to the global reference scale: i.e.

!      ^
!   y  |
!      |
!      |
!      |
!      +---------->
!               x

!   input parameters
!   ----------------
!   Nt       Total number of time steps
!   dt       Time step in percentage of cell traversal; i.e.: dt = c0 dt* / dx*, where dx* = min (Lx/Nx, Ly/Ny)
!   Nout     Number of time steps between outputing properties
!   Nss      Number of time steps until steady state conditions are reached
!   Nx, Ny   Number of computational cells in the x and y directions (regular grid)
!   Lx, Ly   Simulation lengths in dimensionless form: (Lx, Ly) = (Lx*, Ly*) / L0
!              for most simulations, choose min (Lx, Ly) = 1.0
!              when a symmetry condition exists (BC type = 0), choose min (Lx, Ly) = 0.5
!   Xi       Parameter that is approximately equal to the number of particles per cell (~10 for typical problems)
!   Nens     Number of independent statistical ensembles
!   disp     Option for the type of simulation progress information displayed on screen:
!              (1) display simulation information and endpoint only
!              (2) display after each statistical ensemble is finished
!              (3) display after each recorded data point (every Nout steps)
!   Kn       Knudsen number based on the system length scale: Kn = l0 / L0
!   omega    Temperature coefficient of viscosity for the variable-hard-sphere (VHS) collision model ( 1/2 <= omega <= 1 )
!              hard sphere gas:    omega = 0.5
!              Maxwell molecules:  omega = 1.0
!              Argon:              omega = 0.81
!              Helium:             omega = 0.66, etc.
!   epsilon  Departure from equilibrium, which scales all the boundary conditions and streamwise pressure/temperature gradient
!              driving forces.  Note: it is important that the fundamental (largest) nonhomogeniety in the problem (i.e. stream-
!              wise gradients: kappa_P, kappa_T; or boundary conditions: uy*_, uz*_, T*_; discussed below) be chosen to be unity
!              in absolute value.  Thus the departure from equilibrium is to be independently controlled via epsilon without 
!              affecting the overall configuration of the problem.  Following this procedure rigorously is also important to ensure
!              that the parameter Xi (above) is the same order of magnitude of particles in simulation.
!   kappa_P  Dimensionless pressure gradient in the streamwise (z) direction: Pz = - ( L0 / P0 ) ( dP / dz ) / epsilon
!              this is the driving force for Poiseulle flow
!   kappa_T  Dimensionless temperature gradient in the streamwise (z) direction: Tz = ( L0 / T0 ) ( dT / dz ) / epsilon
!              this is the driving force for thermal creep
!   b1,...   Subscripts refering to the boundary condtions corresponding to the geometry (diagrammed above)
!   type_    Boundary condition type:
!              (0) specular
!              (1) diffusely-reflecting
!              (2) Maxwell's accomodation model
!   alpha_   Accomodation coefficient: the probability ( 0.0 < alpha < 1.0 ) of diffuse reflection
!   ux*_     Dimensionless boundary velocity in the x-direction (see above definition)
!   uy*_     Dimensionless boundary velocity in the y-direction (see above definition)
!   uz*_     Dimensionless boundary velocity in the z-direction (see above definition)
!   T*_      Dimensionless boundary temperature (see above definition)

!     Note: the program will remove any nonzero normal velocity components.


!----------------------------------------------------------------------------------------------------------------------------------
! OUTPUT FILES:
!----------------------------------------------------------------------------------------------------------------------------------

!   spatial coordinates ( file name = "xy.dat" )
!   --------------------------------------------
!   x_1   y_1
!   x_2   y_1
!   .
!   .
!   .
!   x_Nx  y_1
!   x_1   y_2
!   x_2   y_2
!   .
!   .
!   .
!   x_Nx  y_2
!   .
!   .
!   .
!   x_1   y_Ny
!   x_2   y_Ny
!   .
!   .
!   .
!   x_Nx  y_Ny

!     There are a total of Nx Ny cells, and the entries of the output file ("output.dat", below) correspond to the cell-centroid
!     positions in the spatial coordinate file ("xy.dat", above).

!   cell-based hydrodynamic properties ( file name = "output.dat" )
!   ---------------------------------------------------------------
!   rho  ux  uy  uz  Pxx  Pxy  Pyy  Pyz  Pzz  Pzx  T  qx  qy  qz

!     if Nt <= Nss in the input file, there will be Nt/Nout sets of Nx Ny values corresponding to each transient time step
!     if Nt > Nss in the input file, there will be a single set of Nx Ny values for the steady state averaged properties


!----------------------------------------------------------------------------------------------------------------------------------
! RATIO OF UNIFORMS BOUNDING FACTORS:     ( file name = "bounds.inp" )
!----------------------------------------------------------------------------------------------------------------------------------

!   The bounding factors are used for sampling the particles created at the boundaries due to advection.  These factors are
!   dynamically tested (when r_Y > 0) and updated when they are exceeded, and each instance will be accompanied by an on-screen
!   warning.  If the simulation finishes normally, the "bounds" file will be rewritten with the updated values.  If a significant
!   number of warning messages are encountered in a simulation, the results may be in error, and the simulation should be rerun.

!   bounds file format
!   ------------------
!   Y_a, Y_bn, Y_bt, r_Y

!   typical values
!   --------------
!   1.0, 1.0, 1.0, 0.01

!   bounding factors
!   ----------------
!   Y_a      Bounding factor for the transformed dependent variable
!   Y_bn     Bounding factor for the transformed particle velocity in the normal (x) direction
!   Y_bt     Bounding factor for the transformed particle velocity in the tangential (y,z) directions
!   r_Y      Percentage of oversampling (0 <= r_Y << 1): the bounds are further modified by a small percentage in order to test the
!              validity of the bounding factors.  For linearized conditions (epsillon << 1) this can be set to zero if all bounding
!              factors (above) are set to unity.  Otherwise, set to a small percentage.


!----------------------------------------------------------------------------------------------------------------------------------
! NUMERICAL PARAMETERS USED IN THE CODE:
!----------------------------------------------------------------------------------------------------------------------------------

!   There are several computational parameters hardwired into the code listing, which can be found within the module "numerical
!   constants".  While the default values should be appropriate for most platforms and typical simulations, they can be modified
!   (if desired) to accomodate special situations.  When these values are exceeded, the user is alerted (by an appropriate error
!   message on the screen), and the simulation stops.

!   computational parameters
!   ------------------------
!   Nmax      Maximum total number of particles in simulation
!   Ncmax     Maximum number of particles contained within any computational cell
!   Nbmax     Maximum number of particles to strike a boundary within a time step


!----------------------------------------------------------------------------------------------------------------------------------
! REFERENCES:
!----------------------------------------------------------------------------------------------------------------------------------

!   [1]  G. A. Radtke, N. G. Hadjiconstantinou, "W. Wagner, Low-noise Monte Carlo simulation of the variable hard sphere gas,"
!        Physics of Fluids, accepted.
!   [2]  W. Wagner, "Deviational particle Monte Carlo for the Boltzmann equation," Monte Carlo Methods and Applications, v 4(3),
!        191-268, 2008.


!==================================================================================================================================
! MODULES
!==================================================================================================================================

! NUMERICAL CONSTANTS
  module numerical_constants
    real(8), parameter :: pi = 3.141592653589793d0
    real(8), parameter :: srpi = 1.772453850905516d0
    real(8), parameter :: eps1 = 1.d-14

!   computational parameters
    integer, parameter :: Nmax = 1000000
    integer, parameter :: Ncmax = 1000
    integer, parameter :: Nbmax = 1000

  end module numerical_constants

!----------------------------------------------------------------------------------------------------------------------------------
! RATIO OF UNIFORMS SAMPLING BOUND CONSTANTS
  module sampling_bound_constants

!   bounds for c_n fMB
    real(8), parameter :: an   = 0.428881942480353d0
    real(8), parameter :: bn_n = 1.344250845932327d0
    real(8), parameter :: bn_s = 0.347897716471792d0

!   bounds for c_n c_n fMB
    real(8), parameter :: ann   = 0.367879441171442d0
    real(8), parameter :: bnn_n = 2.422185882560844d0
    real(8), parameter :: bnn_s = 0.298414096849802d0

!   bounds for c_n c_s fMB
    real(8), parameter :: ans   = 0.183939720585721d0
    real(8), parameter :: bns_n = 0.576524913984314d0
    real(8), parameter :: bns_t = 0.149207048424901d0

!   bounds for c2 c_n fMB
    real(8), parameter :: a2n   = 0.409916278941860d0
    real(8), parameter :: b2n_n = 4.688803555515950d0
    real(8), parameter :: b2n_s = 1.213481884638469d0

  end module sampling_bound_constants

!----------------------------------------------------------------------------------------------------------------------------------
! GLOBAL DATA TYPES
  module types

!   data type for boundary conditions
    type TBC
      integer, dimension(:), allocatable :: type
      integer, dimension(:,:), allocatable :: norm
      integer, dimension(:,:,:), allocatable :: tr, Itr
      real(8), dimension(:), allocatable :: p, c, n, a, Ns, na, ca, dS
      real(8), dimension(:,:), allocatable :: u, b, x, dx
    end type TBC

!   data type for boundary element based particle cross referencing
    type Tbref
      integer, dimension(:,:), allocatable :: N
      integer, dimension(:,:,:), allocatable :: ref
    end type Tbref

!   data type for cell based particle cross referencing
    type Tcref
      logical, dimension(:,:), allocatable :: G
      integer :: DS
      integer, dimension(0:1) :: NG
      integer, dimension(:,:), allocatable :: N, idx, Gidx
      integer, dimension(:,:,:), allocatable :: ref, Glist
      real(8) :: M1max, M1net
      real(8), dimension(:,:), allocatable :: x, M1
    end type Tcref

!   data type for Maxwell-Boltzmann distribution
    type TfMB
      real(8) :: n, c
      real(8), dimension(3) :: u
    end type TfMB

!   data type for numerical parameters for bounding sampled distributions
    type Tbounds
      real(8) :: Y_a, r_Y
      real(8), dimension(2) :: Y_b
    end type Tbounds

!   data type for particle data
    type Tpart
      logical, dimension(:,:), allocatable :: del
      integer, dimension(0:1) :: N
      integer, dimension(:,:), allocatable :: cell
      real(8), dimension(:,:,:), allocatable :: c, x
    end type Tpart

!   data type for property data
    type Tprop
      real(8), dimension(:), allocatable :: n, T
      real(8), dimension(:,:), allocatable :: u, p, q
    end type Tprop

!   data type for simulation parameters
    type Tsim
      logical :: force
      integer :: Nt, Nout, Nss, Nens, disp, Nb
      integer, dimension(2) :: Nc
      real(8) :: dt, Xi, Kn, w, eps, af, Nf, Pz, Tz
      real(8), dimension(2) :: L, dx
      real(8), dimension(3) :: bf
    end type Tsim

!   data type for time step
    type Ttime
      real(8) :: t, tc, dtc
    end type Ttime
  end module types


!==================================================================================================================================
! MAIN PROGRAM
!==================================================================================================================================

  program LVDSMC
    use numerical_constants
    use types
    implicit none

!   variable declarations
    integer :: step, ens
    real(8), dimension(:,:,:), allocatable :: data
    type(Tbref) :: bref
    type(TBC) :: BC
    type(Tcref) :: cref
    type(TfMB) :: f0
    type(Tbounds) :: bounds
    type(Tpart) :: part
    type(Tprop) :: trans, steady
    type(Tsim) :: sim
    type(Ttime) :: time

!   read input data
    call input_data (BC, bref, cref, data, f0, part, trans, steady, sim, time)

!   perform simulation over all ensembles
    do ens=1,sim%Nens

!     initialize state
      call initialize_sim (cref, f0, bounds, part, sim, time)

!     time steps
      do step=1,sim%Nt

!       perform time steps
        time%t = step*sim%dt
        call time_step (BC, bref, f0, bounds, part, trans, steady, sim, step, time)

!       update properties
        call properties (f0, part, trans, sim)
        if (sim%Nt <= sim%Nss) then

!         record transient properties
          if (mod(step,sim%Nout)==0) call record_prop_trans (step/sim%Nout, product(sim%Nc), data, trans)
        end if
        if ( (sim%disp>2) .AND. (mod(step,sim%Nout)==0) ) call display_rec (step, sim%Nt, sum(part%N))
        if (step >= sim%Nss) then

!         record data for steady state properties
          call record_prop_steady (trans, steady, sim)
        end if
      end do

!     display time step after each ensemble is finished
      if (sim%disp>1) call display_ens (ens, sim%Nens, sum(part%N), sim%disp)
    end do

!   output ensemble averaged data
    call display_out (sum(part%N))
    if (sim%Nt>sim%Nss) then
      call output_prop_steady (steady, sim)
    else
      call output_prop_trans (data, sim)
    end if

!   end simulation
    call end_sim (sum(part%N), bounds)

  contains
!----------------------------------------------------------------------------------------------------------------------------------

!   ALLOCATE DATA STORAGE ARRAYS
    subroutine allocate_arrays (BC, bref, cref, data, part, trans, steady, sim, time)
      implicit none
      integer :: Ncell, Ndt, i, j
      real(8), dimension(:,:,:), allocatable :: data
      type(TBC) :: BC
      type(Tbref) :: bref
      type(Tcref) :: cref
      type(Tpart) :: part
      type(Tprop) :: trans, steady
      type(Tsim) :: sim
      type(Ttime) :: time

!     number of cells
      Ncell = product(sim%Nc)

!     allocate arrays
      allocate ( BC%type(sim%Nb), BC%norm(sim%Nb,3), BC%p(sim%Nb), BC%c(sim%Nb), BC%n(sim%Nb), BC%a(sim%Nb), BC%Ns(sim%Nb),       &
                   BC%na(sim%Nb), BC%ca(sim%Nb), BC%x(sim%Nb,2), BC%dx(sim%Nb,2), BC%u(sim%Nb,3), BC%b(sim%Nb,3),                 &
                                                                             BC%tr(sim%Nb,3,3), BC%Itr(sim%Nb,3,3), BC%dS(sim%Nb) )
      allocate (bref%N(0:1,sim%Nb), bref%ref(0:1,sim%Nb,Nbmax))
      allocate ( cref%N(0:1,Ncell), cref%idx(0:1,Nmax), cref%ref(0:1,Ncell,Ncmax), cref%x(Ncell,2), cref%M1(0:1,Nmax),            &
                                                                    cref%G(0:1,Nmax), cref%Gidx(0:1,Nmax), cref%Glist(0:1,1,Nmax) )
      allocate ( part%del(0:1,Nmax), part%x(0:1,Nmax,2), part%c(0:1,Nmax,3), part%cell(0:1,Nmax))
      allocate ( trans%n(Ncell), trans%u(Ncell,3), trans%T(Ncell), trans%p(Ncell,6), trans%q(Ncell,3) )
      allocate ( steady%n(Ncell), steady%u(Ncell,3), steady%T(Ncell), steady%p(Ncell,6), steady%q(Ncell,3) )

!     data structure for ensemble averaged transient data
      if ( sim%Nt <= sim%Nss ) then
        Ndt = sim%Nt / sim%Nout
        allocate ( data(Ncell,Ndt,14) )
        do i = 1, Ncell
          do j = 1, Ndt
            data (i,j,:) = (/0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0,0.d0/)
          end do
        end do
      end if

!     zero out steady properties arrays
      do i = 1, Ncell
        steady%n(i) = 0.d0
        steady%u(i,:) = (/0.d0,0.d0,0.d0/)
        steady%T(i) = 0.d0
        steady%p(i,:) = (/0.d0,0.d0,0.d0,0.d0,0.d0,0.d0/)
        steady%q(i,:) = (/0.d0,0.d0,0.d0/)
      end do

    end subroutine allocate_arrays
!----------------------------------------------------------------------------------------------------------------------------------

!   DRAWS TWO NORMALLY DISTRIBUTED RANDOM VARIABLES USING THE POLAR FORM OF THE BOX MULLER TRANSFORMATION
    subroutine box_Muller (x,y)
      implicit none
      real(8) :: x, y, z

      do
        call random_number(x)
        call random_number(y)
        x=2*x-1
        y=2*y-1
        z=x*x+y*y
        if (z<1.d0) exit
      end do
      z=sqrt((-2*log(z))/z)
      x=sqrt(0.5d0)*x*z
      y=sqrt(0.5d0)*y*z

    end subroutine box_Muller
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY END TIME STAMP
    subroutine display_end (Np)
      implicit none
      integer :: Np
      real :: tcpu

      call cpu_time(tcpu)
      write (0, '(a1,es9.3,i9,a6)') '[',tcpu,Np,']: end'
      write (0, '(a48)') '================================================'
      flush (0)

    end subroutine display_end
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY END TIME STAMP
    subroutine display_ens (ens, Nens, Np, disp)
      implicit none
      integer :: ens, Nens, Np, disp
      real :: tcpu

      call cpu_time(tcpu)
      if (disp>2) write (0, '(a48)') '------------------------------------------------'
      write (0, '(a1,es9.3,i9,a11,i7,a3,i7)') '[',tcpu,Np,']: ensemble',ens,' of',Nens
      if (disp>2) write (0, '(a48)') '------------------------------------------------'
      flush (0)

    end subroutine display_ens
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY INITIAL TIME STAMP
    subroutine display_init (Np)
      implicit none
      integer :: Np
      real :: tcpu

      call cpu_time(tcpu)
      write (0, '(a1,es9.3,i9,a24)') '[',tcpu,Np,']: initial condition set'
      write (0, '(a48)') '------------------------------------------------'
      flush (0)

    end subroutine display_init
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY TIME STAMP FOR WRITING THE OUTPUT FILE
    subroutine display_out (Np)
      implicit none
      integer :: Np
      real :: tcpu

      call cpu_time(tcpu)
      write (0, '(a1,es9.3,i9,a22)') '[',tcpu,Np,']: writing output file'
      flush (0)

    end subroutine display_out
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY INITIAL SIMULATION PARAMETERS
    subroutine display_sim (BC, sim)
      implicit none
      integer :: i
      real :: tcpu
      type(TBC) :: BC
      type(Tsim) :: sim

!     initialize cpu time
      call cpu_time(tcpu)

!     header
      write (0, *)
      write (0, *)
      write (0, '(a48)') '================================================'
      write (0, '(a29)') 'LVDSMC_VHS_2D.f90 version 1.0'
      write (0, '(a42)') 'low-noise 2D Boltzmann equation simulation'
      write (0, '(a36)') 'variable hard sphere collision model'
      write (0, '(a26)') '  written by: Gregg Radtke'
      write (0, '(a48)') '------------------------------------------------'
      write (0, *)

!     general simulation parameters
      write (0, '(a22)') 'simulation parameters:'
      write (0, '(a48)') '================================================'
      write (0, '(a28,es10.3)') 'Knudsen number:             ', sim%Kn
      write (0, '(a28,es10.3)') 'Temperature coefficient:    ', sim%w
      write (0, '(a28,es10.3)') 'Departure from equilibrium: ', sim%eps
      write (0, '(a28,es10.3)') 'Pressure gradient (z):      ', sim%Pz
      write (0, '(a28,es10.3)') 'Temperature gradient (z):   ', sim%Tz
      write (0, '(a28,es10.3)') 'Simulation domain size (x): ', sim%L(1)
      write (0, '(a28,es10.3)') 'Simulation domain size (y): ', sim%L(2)
      write (0, '(a28,es10.3)') 'Approx. particles per cell: ', sim%Xi
      write (0, '(a28,es10.3)') 'Advective time (final):     ', sim%dt * sim%Nt
      write (0, '(a28,es10.3)') 'Advective time (steady):    ', sim%dt * sim%Nss
      write (0, '(a29,i8)') 'Total number of cells:       ', product(sim%Nc)
      write (0, '(a29,i8)') 'Total number of boundaries:  ', sim%Nb

!     averaging method
      if ( sim%Nt > sim%Nss ) then
        write (0, '(a42)') 'Steady state averaging (steady simulation)'
      else
        write (0, '(a41)') 'Ensemble averaging (transient simulation)'
      end if
      write (0, '(a29,i8)') 'Number of ensembles:         ', sim%Nens
      write (0, '(a48)') '------------------------------------------------'
      write (0, *)

!     boundary condtitions
      write (0, '(a20)') 'boundary conditions:'
      write (0, '(a56)') '========================================================'
      write (0, '(a51)') '        x      y      alpha  ux     uy     uz     T'
      write (0, '(a56)') '========================================================'
      do i = 1, sim%Nb
        select case ( BC%type(i) )
        case (0)
          write (0, '(i6,2f7.3,a7,4f7.3)') i, BC%x(i,:)+BC%dx(i,:)/2, '  0.000', BC%u(i,:), BC%c(i)
        case (1)
          write (0, '(i6,2f7.3,a7,4f7.3)') i, BC%x(i,:)+BC%dx(i,:)/2, '  1.000', BC%u(i,:), BC%c(i)
        case (2)
          write (0, '(i6,2f7.3,5f7.3)') i, BC%x(i,:)+BC%dx(i,:)/2, BC%p(i), BC%u(i,:), BC%c(i)
        case default
          write (0, '(a56)') '--------------------------------------------------------'
          stop 'ERROR: incorrect boundary condition type specified'
        end select
      end do
      write (0, '(a56)') '--------------------------------------------------------'

!     simulation progress header
      write (0, *)
      write (0, '(a41)') '[time,    particles]: simulation progress'
      write (0, '(a48)') '================================================'
      flush (0)

    end subroutine display_sim
!----------------------------------------------------------------------------------------------------------------------------------

!   DISPLAY TIME STAMP EVERY Nout STEPS
    subroutine display_rec (step, Nt, Np)
      implicit none
      integer :: step, Np, Nt
      real :: tcpu
      call cpu_time(tcpu)
      write (0, '(a1,es9.3,i9,a7,i9,a3,i9)') '[',tcpu,Np,']: step',step,' of',Nt
      flush (0)

    end subroutine display_rec
!----------------------------------------------------------------------------------------------------------------------------------

!   END SIMULATION
    subroutine end_sim (Np, bounds)
      integer :: Np
      type(Tbounds) :: bounds

!     update bounds factors
      open (2, file='bounds.inp')
      rewind (2)
      write (2,*) bounds%Y_a, bounds%Y_b, real(bounds%r_Y - 1.d0)
      close (2)

!     display end time stamp
      call display_end (Np)

    end subroutine end_sim
!----------------------------------------------------------------------------------------------------------------------------------

!   EVALUATE MAXWELLIAN
    function eval_fMB (c, fMB)
      implicit none
      real(8) :: eval_fMB
      real(8), dimension(3) :: c
      type(TfMB) :: fMB

      eval_fMB = fMB%n / (srpi*fMB%c)**3 * exp ( - dot_product( c-fMB%u, c-fMB%u ) / fMB%c**2 )

    end function eval_fMB
!----------------------------------------------------------------------------------------------------------------------------------

!   INITIALIZE SIMULATION
    subroutine initialize_sim (cref, f0, bounds, part, sim, time)
      implicit none
      integer :: i
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tbounds) :: bounds
      type(Tpart) :: part
      type(Tsim) :: sim
      type(Ttime) :: time

!     Reset collision time counters
      time%tc = 0.d0
      time%dtc = -1.d0

!     reset number of simulated particles
      part%N = (/0,0/)

!     initialize collision mass conservation variable
      cref%DS = 0

!     initialization time stamp
      if (sim%disp>2) call display_init (sum(part%N))

    end subroutine initialize_sim
!----------------------------------------------------------------------------------------------------------------------------------

!   READ INPUT DATA AND FIX SIMULATION PARAMETERS
    subroutine input_data (BC, bref, cref, data, f0, part, trans, steady, sim, time)
      use sampling_bound_constants
      implicit none
      integer :: i, cell
      real(8), dimension(:,:,:), allocatable :: data
      type(TBC) :: BC
      type(Tbref) :: bref
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tpart) :: part
      type(Tprop) :: trans, steady
      type(Tsim) :: sim
      type(Ttime) :: time

!     open input file
      open (1,file='input.inp',status='old')

!     read input data header
      read (1,*) sim%Nt, sim%dt, sim%Nout, sim%Nss
      read (1,*) sim%Nc, sim%L, sim%Xi, sim%Nens, sim%disp
      read (1,*) sim%Kn, sim%w, sim%eps, sim%Pz, sim%Tz

!     total number of boundaries
      sim%Nb = 2*sim%Nc(1) + 2*sim%Nc(2)

!     allocate cell and particle arrays
      call allocate_arrays (BC, bref, cref, data, part, trans, steady, sim, time)

!     read boundary conditions
      do i = 1, sim%Nb
        read (1,*) BC%type(i), BC%p(i), BC%u(i,:), BC%c(i)
      end do
      close (1)

!     read bound factor data
      open (2,file='bounds.inp',status='old')
      read (2,*) bounds%Y_a, bounds%Y_b, bounds%r_Y
      close (2)

!     reference equilibrium distribution
      f0%n = 1.d0
      f0%u = (/0.d0,0.d0,0.d0/)
      f0%c = 1.d0

!     set geometrical parameters
      call set_geometry (BC, cref, sim)

!     rescale time step to the advective time scale
      sim%dt = sim%dx(1)*sim%dt/f0%c

!     display simulation parameters
      call display_sim (BC, sim)

!     convert perturbed boundary conditions to true boundary conditions
      do i = 1, sim%Nb
        BC%u(i,:) = sim%eps*BC%u(i,:)
        BC%c(i) = sqrt(1 + sim%eps*BC%c(i))
      end do

!     set boundary densities for reflections
      do i = 1, sim%Nb
        BC%n(i) = f0%n * (f0%c/BC%c(i))
      end do

!     set force logical variable
      if ( abs(sim%Pz) + abs(sim%Tz) > eps1 ) then
        sim%force = .true.
      else
        sim%force = .false.
      end if

!     create spatial coordinate data file
      open (23, file='xy.dat', status='unknown')
      do cell = 1, product(sim%Nc)
        write (23,'(3es13.5)') cref%x(cell,1)+sim%dx(1)/2, cref%x(cell,2)+sim%dx(2)/2
      end do
      close (23)

!     calculate Neff, which is scaled by the perturbation parameter and reference density
      sim%Xi = sim%eps/sim%Xi

!     convert perturbed body forces to true body forces
      sim%Pz = sim%Pz * sim%eps
      sim%Tz = sim%Tz * sim%eps

!     set sampling bounds for body force generation term
      call set_force_bounds (f0, sim)

!     convert Knudsen number to dimensionless prefactor for collision integral
      sim%Kn = (8.d0/pi)**sim%w / (16.d0*srpi) / sim%Kn

!     convert temperature exponent of viscosity to exponent of relative velocity in collision rate
      sim%w = 2*(1 - sim%w)

!     force accomodation coefficients for specular boundary condtions to be zero
      do i = 1, sim%Nb
        if (BC%type(i)==0) BC%p(i) = 0.d0
      end do

!     transform boundary velocity to local coordinates
      do i = 1, sim%Nb
        BC%u(i,:) = matmul( BC%tr(i,:,:), BC%u(i,:) )
      end do

!     set constants for sampling boundary generation term
      call set_advection_bounds (BC, f0, sim)

!     convert bound factors
      bounds%r_Y  = 1 + bounds%r_Y

!     reset seed for random number generator
      call reset_seed

    end subroutine input_data
!----------------------------------------------------------------------------------------------------------------------------------

!   OUTPUT ENSEMBLE AVERAGED DATA
    subroutine output_prop_trans (data, sim)
      implicit none
      integer :: Ndt, i, j
      real(8) :: n, T, u2, eps
      real(8), dimension(3) :: u, q
      real(8), dimension(6) :: p
      real(8), dimension(:,:,:), allocatable :: data
      type(Tsim) :: sim

!     number of time steps
      Ndt = sim%Nt / sim%Nout

!     perturbation parameter
      eps = sim%eps

!     perform ensemble averaging on the moments of the velocity distribution
      data = data/sim%Nens

!     open data output file
      open (3, file='output.dat', status='unknown')

!     output ensemble averaged values
      do j = 1, Ndt
      do i = 1, product(sim%Nc)

!       construct dimensionless hydrodynamic properties from velocity moments
        n = sim%Xi * data(i,j,1)
        u = sim%Xi / (1.d0+n) * data(i,j,2:4)
        u2 = dot_product(u,u)
        p(1) = 2.d0*( sim%Xi*data(i,j,5) - (1.d0+n)*u(1)**2 )
        p(2) = 2.d0*( sim%Xi*data(i,j,6) - (1.d0+n)*u(1)*u(2) )
        p(3) = 2.d0*( sim%Xi*data(i,j,7) - (1.d0+n)*u(2)**2 )
        p(4) = 2.d0*( sim%Xi*data(i,j,8) - (1.d0+n)*u(2)*u(3) )
        p(5) = 2.d0*( sim%Xi*data(i,j,9) - (1.d0+n)*u(3)**2 )
        p(6) = 2.d0*( sim%Xi*data(i,j,10) - (1.d0+n)*u(3)*u(1) )
        T = ( sim%Xi*data(i,j,11)/1.5d0 - n)/(1.d0+n) - u2/1.5d0
        q(1) = sim%Xi*data(i,j,12) - u(1) - p(1)*u(1) - p(2)*u(2) - p(6)*u(3) - (1.d0+n)*u(1)*(1.5d0*(1.d0+T)+u2)
        q(2) = sim%Xi*data(i,j,13) - u(2) - p(2)*u(1) - p(3)*u(2) - p(4)*u(3) - (1.d0+n)*u(2)*(1.5d0*(1.d0+T)+u2)
        q(3) = sim%Xi*data(i,j,14) - u(3) - p(6)*u(1) - p(4)*u(2) - p(5)*u(3) - (1.d0+n)*u(3)*(1.5d0*(1.d0+T)+u2)

!       scale properties based on eps
        n = n/eps
        u = u/eps
        p = p/eps
        T = T/eps
        q = q/eps

!       write values to file
        write (3,'(14es13.5)') n, u, p, T, q

      end do
      end do

!     close data output file
      close (3)

    end subroutine output_prop_trans
!----------------------------------------------------------------------------------------------------------------------------------

!   OUTPUT STEADY ENSEMBLE AVERAGED DATA
    subroutine output_prop_steady (steady, sim)
      implicit none
      integer :: Navg, i
      real(8) :: n, T, u2, eps
      real(8), dimension(3) :: u, q
      real(8), dimension(6) :: p
      type(Tprop) :: steady
      type(Tsim) :: sim

!     number of points averaged over
      Navg = sim%Nens*(sim%Nt+1-sim%Nss)

!     perturbation parameter
      eps = sim%eps

!     open data output file
      open (3, file='output.dat', status='unknown')

!     output ensemble averaged values
      do i = 1, product(sim%Nc)

!       perform ensemble averaging
        n = steady%n(i)/Navg
        u = steady%u(i,:)/Navg
        p = steady%p(i,:)/Navg
        T = steady%T(i)/Navg
        q = steady%q(i,:)/Navg

!       construct dimensionless hydrodynamic properties from velocity moments
        n = sim%Xi * n
        u = sim%Xi/(1.d0+n) * u
        u2 = dot_product(u,u)
        p(1) = 2.d0*( sim%Xi*p(1) - (1.d0+n)*u(1)**2 )
        p(2) = 2.d0*( sim%Xi*p(2) - (1.d0+n)*u(1)*u(2) )
        p(3) = 2.d0*( sim%Xi*p(3) - (1.d0+n)*u(2)**2 )
        p(4) = 2.d0*( sim%Xi*p(4) - (1.d0+n)*u(2)*u(3) )
        p(5) = 2.d0*( sim%Xi*p(5) - (1.d0+n)*u(3)**2 )
        p(6) = 2.d0*( sim%Xi*p(6) - (1.d0+n)*u(3)*u(1) )
        T = ( sim%Xi*T/1.5d0 - n)/(1.d0+n) - u2/1.5d0
        q(1) = sim%Xi*q(1) - u(1) - p(1)*u(1) - p(2)*u(2) - p(6)*u(3) - (1.d0+n)*u(1)*(1.5d0*(1.d0+T)+u2)
        q(2) = sim%Xi*q(2) - u(2) - p(2)*u(1) - p(3)*u(2) - p(4)*u(3) - (1.d0+n)*u(2)*(1.5d0*(1.d0+T)+u2)
        q(3) = sim%Xi*q(3) - u(3) - p(6)*u(1) - p(4)*u(2) - p(5)*u(3) - (1.d0+n)*u(3)*(1.5d0*(1.d0+T)+u2)

!       scale properties based on eps
        n = n/eps
        u = u/eps
        p = p/eps
        T = T/eps
        q = q/eps

!       write values to file
        write (3,'(14es13.5)') n, u, p, T, q
      end do

!     close data output file
      close (3)

    end subroutine output_prop_steady
!----------------------------------------------------------------------------------------------------------------------------------

!   UPDATE PROPERTIES
    subroutine properties (f0, part, trans, sim)
      implicit none
      integer :: i, p, s
      real(8) :: c2
      real(8), dimension(3) :: c
      type(TfMB) :: f0
      type(Tpart) :: part
      type(Tprop) :: trans
      type(Tsim) :: sim

!     zero out properties
      do i = 1, product(sim%Nc)
        trans%n(i) = 0.d0
        trans%u(i,:) = (/0.d0,0.d0,0.d0/)
        trans%T(i) = 0.d0
        trans%p(i,:) = (/0.d0,0.d0,0.d0,0.d0,0.d0,0.d0/)
        trans%q(i,:) = (/0.d0,0.d0,0.d0/)
      end do

!     sum the moments corresponding to density, velocity, pressure tensor, temperature, and heat flux

!     positive particles
      do p=1,part%N(0)
        i = part%cell(0,p)
        c = (part%c(0,p,:) - f0%u)/f0%c
        trans%n(i) = trans%n(i) - 1
        trans%u(i,:) = trans%u(i,:) - c
        trans%p(i,1) = trans%p(i,1) - c(1)**2
        trans%p(i,2) = trans%p(i,2) - c(1)*c(2)
        trans%p(i,3) = trans%p(i,3) - c(2)**2
        trans%p(i,4) = trans%p(i,4) - c(2)*c(3)
        trans%p(i,5) = trans%p(i,5) - c(3)**2
        trans%p(i,6) = trans%p(i,6) - c(3)*c(1)
        c2 = dot_product(c,c)
        trans%T(i) = trans%T(i) - c2
        trans%q(i,:) = trans%q(i,:) - part%c(0,p,:)*c2
      end do

!     negative particles
      do p=1,part%N(1)
        i = part%cell(1,p)
        c = (part%c(1,p,:) - f0%u)/f0%c
        trans%n(i) = trans%n(i) + 1
        trans%u(i,:) = trans%u(i,:) + c
        trans%p(i,1) = trans%p(i,1) + c(1)**2
        trans%p(i,2) = trans%p(i,2) + c(1)*c(2)
        trans%p(i,3) = trans%p(i,3) + c(2)**2
        trans%p(i,4) = trans%p(i,4) + c(2)*c(3)
        trans%p(i,5) = trans%p(i,5) + c(3)**2
        trans%p(i,6) = trans%p(i,6) + c(3)*c(1)
        c2 = dot_product(c,c)
        trans%T(i) = trans%T(i) + c2
        trans%q(i,:) = trans%q(i,:) + part%c(1,p,:)*c2
      end do

    end subroutine properties
!----------------------------------------------------------------------------------------------------------------------------------

!   PRODUCE A RANDOM INTEGER FROM 1 TO N
    function rand_int(N)
      implicit none
      integer :: N, rand_int
      real(8) :: r

      call random_number(r)
      rand_int = int(r*N) + 1

    end function rand_int
!----------------------------------------------------------------------------------------------------------------------------------

!   SAVE DATA TO BINARY FILE FOR ENSEMBLE AVERAGING
    subroutine record_prop_trans (step, Ncell, data, trans)
      implicit none
      integer :: step, Ncell, i
      real(8), dimension(:,:,:) :: data
      type(Tprop) :: trans

      do i = 1, product(sim%Nc)
        data(i,step,1)     = data(i,step,1)     + trans%n(i)
        data(i,step,2:4)   = data(i,step,2:4)   + trans%u(i,:)
        data(i,step,5:10)  = data(i,step,5:10)  + trans%p(i,:)
        data(i,step,11)    = data(i,step,11)    + trans%T(i)
        data(i,step,12:14) = data(i,step,12:14) + trans%q(i,:)
      end do

    end subroutine record_prop_trans
!----------------------------------------------------------------------------------------------------------------------------------

!   SAVE STEADY STATE PROPERTY DATA FOR ENSEMBLE AVERAGING
    subroutine record_prop_steady (trans, steady, sim)
      implicit none
      integer :: i
      type (Tprop) :: trans, steady
      type (Tsim) :: sim

!     cell-based properties
      do i = 1, product(sim%Nc)
        steady%n(i) = steady%n(i) + trans%n(i)
        steady%u(i,:) = steady%u(i,:) + trans%u(i,:)
        steady%p(i,:) = steady%p(i,:) + trans%p(i,:)
        steady%T(i) = steady%T(i) + trans%T(i)
        steady%q(i,:) = steady%q(i,:) + trans%q(i,:)
      end do

    end subroutine record_prop_steady
!----------------------------------------------------------------------------------------------------------------------------------

!   RESET SEED FOR RANDOM NUMBER GENERATOR TO SYSTEM DATE AND TIME DERIVED VALUE
    subroutine reset_seed
      implicit none
      integer :: i_seed
      integer, dimension(:), allocatable :: a_seed
      integer, dimension(1:8) :: dt_seed

      call random_seed (size = i_seed)
      allocate (a_seed(1:i_seed))
      call random_seed (get = a_seed)
      call date_and_time (values = dt_seed)
      a_seed(i_seed) = dt_seed(8); a_seed(1) = dt_seed(8) * dt_seed(7) * dt_seed(6)
      call random_seed (put = a_seed)
      deallocate (a_seed)

    end subroutine reset_seed
!----------------------------------------------------------------------------------------------------------------------------------

!   PRODUCE A SINGLE SAMPLE FROM A GUASSIAN DISTRIBUTION
    subroutine sample_fMB (c, u, v)
      implicit none
      real(8) :: v, c3
      real(8), dimension(3) :: c, u

!     produce sample
      call box_Muller (c(1),c(2))
      call box_muller (c(3),c3)

!     scale mean and variance
      c = u + v*c

    end subroutine sample_fMB
!----------------------------------------------------------------------------------------------------------------------------------

!   SET BOUNDARY GENERATION BOUNDS BASED ON LINEARIZED CONDITIONS
    subroutine set_advection_bounds (BC, f0, sim)
      use numerical_constants
      use types
      use sampling_bound_constants
      implicit none
      integer :: i
      real(8) :: na, dn, ca, dc, acc
      real(8), dimension(3) :: du
      type(TBC) :: BC
      type(TfMB) :: f0
      type(Tsim) :: sim

      do i=1,sim%Nb

!       for specular boundary elements, prevent particle generation and cycle to next boundary element
        BC%Ns(i) = 1.d0
        if (BC%type(i)==0) then
          BC%Ns(i) = 0.d0
          cycle
        end if

!       compute average properties for generating particles at the boundaries
        na = (f0%n + BC%n(i))/2
        ca = (f0%c + BC%c(i))/2

!       compute the absolute property differences for estimating the sampling bounds
        dn = (BC%n(i) - f0%n)/na
        du = (BC%u(i,:) - f0%u)/ca
        dc = (BC%c(i) - f0%c)/ca

!       compute linearized sampling bounds
        BC%a(i) = ( abs(dn-3.d0*dc)*an + 2.d0*abs(du(1))*ann + 2.d0*abs(du(2))*ans + 2.d0*abs(du(3))*ans + 2.d0*abs(dc)*a2n )**0.4
        BC%b(i,1)   = ( abs(dn-3.d0*dc)*bn_n + 2.d0*abs(du(1))*bnn_n + 2.d0*abs(du(2))*bns_n + 2*abs(du(3))*bns_n                 &
                                                                                                        + 2.d0*abs(dc)*b2n_n )**0.2
        BC%b(i,2)   = ( abs(dn-3.d0*dc)*bn_s + 2.d0*abs(du(1))*bnn_s + 2.d0*abs(du(2))*bns_n + 2*abs(du(3))*bns_t                 &
                                                                                                        + 2.d0*abs(dc)*b2n_s )**0.2
        BC%b(i,3)   = ( abs(dn-3.d0*dc)*bn_s + 2.d0*abs(du(1))*bnn_s + 2.d0*abs(du(2))*bns_t + 2*abs(du(3))*bns_n                 &
                                                                                                        + 2.d0*abs(dc)*b2n_s )**0.2

!       compute number of trial samples per time step dt
        BC%Ns(i) = 10.d0/srpi**3 * na*ca * BC%dS(i)/sim%Xi/product(sim%dx) * BC%a(i)*product(BC%b(i,:)) * BC%Ns(i)

!       adjust for accomodating boundary condtion type
        if (BC%type(i)==2) BC%Ns(i) = BC%Ns(i) * BC%p(i)

!       store mean density and most probable speed
        BC%na(i) = na
        BC%ca(i) = ca
      end do

    end subroutine set_advection_bounds
!----------------------------------------------------------------------------------------------------------------------------------

!   SET BOUNDARY GENERATION BOUNDS BASED ON LINEARIZED CONDITIONS
    subroutine set_force_bounds (f0, sim)
      use numerical_constants
      use types
      use sampling_bound_constants
      implicit none
      type(TfMB) :: f0
      type(Tsim) :: sim

!     set sampling bounds for body force generation term
      sim%af = ( an*abs(sim%Pz) + (2.5d0*an+a2n)*abs(sim%Tz) )**0.4
      sim%bf(1) = ( bn_s*abs(sim%Pz) + (2.5d0*bn_s+b2n_s)*abs(sim%Tz) )**0.2
      sim%bf(2) = ( bn_s*abs(sim%Pz) + (2.5d0*bn_s+b2n_s)*abs(sim%Tz) )**0.2
      sim%bf(3) = ( bn_n*abs(sim%Pz) + (2.5d0*bn_n+b2n_n)*abs(sim%Tz) )**0.2
      sim%Nf = 20.d0 * f0%n/(srpi*f0%c)**3 *sim%af*product(sim%bf) * (sim%dt/2) * product(sim%L)/product(sim%dx)/sim%Xi

    end subroutine set_force_bounds
!----------------------------------------------------------------------------------------------------------------------------------

!   SET GEOMETRICAL PARAMETERS
    subroutine set_geometry (BC, cref, sim)
    implicit none
      integer :: i, j, cell
      integer, dimension(3,3) :: tr1, tr2, tr3, tr4
      type(TBC) :: BC
      type(Tcref) :: cref
      type(Tsim) :: sim

!     total number of boundary conditions
      sim%Nb = 2*sim%Nc(1) + 2*sim%Nc(2)

!     calculate cell spatial disctretization
      sim%dx(1) = sim%L(1)/sim%Nc(1)
      sim%dx(2) = sim%L(2)/sim%Nc(2)

!     set corner coordinates
      do cell = 1, sim%Nc(1)
        cref%x(cell,1) = sim%dx(1)*(cell-1)
      end do

!     set corner coordinates
      do j = 1, sim%Nc(2); do i = 1, sim%Nc(1)
        cell = i + sim%Nc(1)*(j-1)
        cref%x(cell,1) = sim%dx(1)*(i-1)
        cref%x(cell,2) = sim%dx(2)*(j-1)
      end do; end do

!     boundary element corners
      do i = 1, sim%Nc(1)
        BC%x(i,:) = (/ (i-1)*sim%dx(1), 0.d0 /)
        BC%x(2*sim%Nc(1)+sim%Nc(2)+1-i,:) = (/ (i-1)*sim%dx(1), sim%L(2) /)
      end do
      do j = 1, sim%Nc(2)
        BC%x(sim%Nc(1)+j,:) = (/ sim%L(1), (j-1)*sim%dx(2) /)
        BC%x(2*(sim%Nc(1)+sim%Nc(2))+1-j,:) = (/ 0.d0, (j-1)*sim%dx(2) /)
      end do

!     boundary element dimensions
      do i = 1, sim%Nc(1)
        BC%dx(i,:) = (/ sim%dx(1), 0.d0 /)
        BC%dx(2*sim%Nc(1)+sim%Nc(2)+1-i,:) = (/ sim%dx(1), 0.d0 /)
      end do
      do j = 1, sim%Nc(2)
        BC%dx(sim%Nc(1)+j,:) = (/ 0.d0, sim%dx(2) /)
        BC%dx(2*(sim%Nc(1)+sim%Nc(2))+1-j,:) = (/ 0.d0, sim%dx(2) /)
      end do

!     inward surface normals
      do i = 1, sim%Nc(1)
        BC%norm(i,:) = (/ 0, 1, 0 /)
        BC%norm(2*sim%Nc(1)+sim%Nc(2)+1-i,:) = (/ 0, -1, 0 /)
      end do
      do j = 1, sim%Nc(2)
        BC%norm(sim%Nc(1)+j,:) = (/ -1, 0, 0 /)
        BC%norm(2*(sim%Nc(1)+sim%Nc(2))+1-j,:) = (/ 1, 0, 0 /)
      end do

!     remove normal velocity components
      do i = 1, sim%Nc(1)
        BC%u(i,2) = 0.d0
        BC%u(2*sim%Nc(1)+sim%Nc(2)+1-i,2) = 0.d0
      end do
      do j = 1, sim%Nc(2)
        BC%u(sim%Nc(1)+j,1) = 0.d0
        BC%u(2*(sim%Nc(1)+sim%Nc(2))+1-j,1) = 0.d0
      end do

!     transformation matrices
      tr1(1,:) = (/0,1,0/)
      tr1(2,:) = (/-1,0,0/)
      tr1(3,:) = (/0,0,1/)

      tr2(1,:) = (/-1,0,0/)
      tr2(2,:) = (/0,1,0/)
      tr2(3,:) = (/0,0,1/)

      tr3(1,:) = (/0,-1,0/)
      tr3(2,:) = (/1,0,0/)
      tr3(3,:) = (/0,0,1/)

      tr4(1,:) = (/1,0,0/)
      tr4(2,:) = (/0,1,0/)
      tr4(3,:) = (/0,0,1/)

      do i = 1, sim%Nc(1)
        BC%tr(i,:,:) = tr1
        BC%tr(2*sim%Nc(1)+sim%Nc(2)+1-i,:,:) = tr3
      end do
      do j = 1, sim%Nc(2)
        BC%tr(sim%Nc(1)+j,:,:) = tr2
        BC%tr(2*(sim%Nc(1)+sim%Nc(2))+1-j,:,:) = tr4
      end do

!     inverse transformation matrices
      do i = 1, sim%Nc(1)
        BC%Itr(i,:,:) = transpose(tr1)
        BC%Itr(2*sim%Nc(1)+sim%Nc(2)+1-i,:,:) = transpose(tr3)
      end do
      do j = 1, sim%Nc(2)
        BC%Itr(sim%Nc(1)+j,:,:) = transpose(tr2)
        BC%Itr(2*(sim%Nc(1)+sim%Nc(2))+1-j,:,:) = transpose(tr4)
      end do

!     surface area elements
      do i = 1, sim%Nc(1)
        BC%dS(i) = sim%dx(1)
        BC%dS(2*sim%Nc(1)+sim%Nc(2)+1-i) = sim%dx(1)
      end do
      do j = 1, sim%Nc(2)
        BC%dS(sim%Nc(1)+j) = sim%dx(2)
        BC%dS(2*(sim%Nc(1)+sim%Nc(2))+1-j) = sim%dx(2)
      end do

    end subroutine set_geometry
!----------------------------------------------------------------------------------------------------------------------------------

!   SHUFFLE THE ELEMENTS IN A VECTOR OF LENGTH N
    subroutine shuffle (X, N)
      implicit none
      integer :: N, i, j, Xs
      integer, dimension(:) :: X

      do i=1,N
        j = rand_int(N+1-i) + (i-1)
        Xs = X(j)
        X(j) = X(i)
        X(i) = Xs
      end do

    end subroutine shuffle
!----------------------------------------------------------------------------------------------------------------------------------

!   CHOOSE AN INDEX BASED ON PARTICLE SIGN (1/0=positive/negative)
    function sign_index (x)
      implicit none
      integer :: sign_index
      real(8) :: x

      if (x>=0.d0) then
        sign_index = 1
      else
        sign_index = 0
      end if

    end function sign_index
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Nmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a48)') '[',tcpu,Np,']: ERROR: Nmax exceeded--increase Nmax parameter'
      write (0,*)
      flush (0)
      stop

    end subroutine stop_exceed_Nmax
!----------------------------------------------------------------------------------------------------------------------------------

!   ADVANCE SIMULATION BY ONE TIME STEP
!   second-order time splitting similar to Strang's method for DSMC [see Rader, et. al. Phys. Fluids (2006)]
    subroutine time_step (BC, bref, f0, bounds, part, trans, steady, sim, step, time)
      implicit none
      integer :: step
      type (TBC) :: BC
      type (Tbref) :: bref
      type (TfMB) :: f0
      type (Tbounds) :: bounds
      type (Tpart) :: part
      type (Tprop) :: trans, steady
      type (Tsim) :: sim
      type (Ttime) :: time

!     perform intial advection half step
      call advection (BC, bref, f0, bounds, part, trans, sim)

!     perform effective "body force" half step
      if (sim%force) call force (f0, part, sim)

!     perform full collision step
      call collision (cref, f0, part, trans, sim, time)

!     perform effective "body force" half step
      if (sim%force) call force (f0, part, sim)

!     perform final half advection step
      call advection (BC, bref, f0, bounds, part, trans, sim)

    end subroutine time_step
!----------------------------------------------------------------------------------------------------------------------------------

  end program LVDSMC


!==================================================================================================================================
! ADVECTION ROUTINE
!==================================================================================================================================

! ADVECTION ROUTINE FOR HALF TIME STEP
  subroutine advection (BC, bref, f0, bounds, part, trans, sim)
    use types
    use numerical_constants
    implicit none
    logical :: esc
    integer :: i, j, cell, type
    integer, dimension(0:1) :: NG
    real(8) :: dt, r, dtf
    type(TBC) :: BC
    type(Tbref) :: bref
    type(TfMB) :: f0
    type(Tbounds) :: bounds
    type(Tpart) :: part
    type(Tprop) :: trans
    type(Tsim) :: sim

!   time step
    dt = sim%dt/2

!   clear boundary cell index data
    do i = 1, sim%Nb
      bref%N(:,i) = (/0,0/);
    end do

!   advect particles along velocity trajectories
    do i = 0, 1
    do j = 1, part%N(i)
      part%x(i,j,1) = part%x(i,j,1) + dt * part%c(i,j,1)
      part%x(i,j,2) = part%x(i,j,2) + dt * part%c(i,j,2)
      call locate (esc, cell, part%x(i,j,:), part%c(i,j,:), sim)
      part%cell(i,j) = cell

!     enforce boundary condition for escaping particles
      if (esc) then

!       determine boundary condition type
        type = BC%type(cell)
        if (type==2) then
          call random_number(r)
          if (r<BC%p(cell)) then
            type = 1
          else
            type = 0
          end if
        end if
        select case (type)

!       specular reflection
        case (0)
          dtf = dt
          call specular_reflection (i, j, dtf, BC, part, trans, sim)

!       diffuse reflection
        case (1)

!         warn if the maximum number of particles per boundary is exceeded
          if (bref%N(i,cell) >= Nbmax) call stop_exceed_Nbmax (sum(part%N(:)))

!         create lists of positive and negative particles for cancelling and reflection
          bref%N(i,cell) = bref%N(i,cell) + 1
          bref%ref(i,cell,bref%N(i,cell)) = j
        end select

!     update particle cell location if it remains within the simulation domain
      else
        part%cell(i,j) = cell
      end if
    end do
    end do

!   mark pairs of positive and negative particles from diffuse reflections for deletion
    dtf = dt
    call boundary_cancel (dtf, BC, bref, part, sim)

!   removed deleted particles from simulation
    do i = 0, 1
    do j = part%N(i), 1, -1
      if (part%del(i,j)) call remove(i, j, part)
    end do
    end do

!   generate particles at the boundary
    do i = 1, sim%Nb

!     skip routine for specular reflections
      if (BC%type(i)==0) then

!     advection particle generation routine with mass conservation
      else
        call generate_boundary (i, dt, 0.5d0, BC, f0, bounds, part, trans, sim, NG, .false.)
        j = NG(0)
        NG(0) = NG(1)
        NG(1) = j
        call generate_boundary (i, dt, 0.5d0, BC, f0, bounds, part, trans, sim, NG, .true.)
      end if
    end do

  contains
!----------------------------------------------------------------------------------------------------------------------------------

!   ADD PARTICLE TO SIMULATION
    subroutine add (c, x, s, cell, part)
      implicit none
      integer :: s, cell
      real(8), dimension(2) :: x
      real(8), dimension(3) :: c
      type(Tpart) :: part

!     warn if the maximum number of particles is exceeded
      if (maxval(part%N)>=Nmax) call stop_exceed_Nmax (sum(part%N))

!     add new velocity sample
      part%N(s) = part%N(s) + 1
      part%c(s, part%N(s), :) = c
      part%x(s, part%N(s), :) = x
      part%cell(s, part%N(s)) = cell
      part%del(s, part%N(s)) = .false.

    end subroutine add
!----------------------------------------------------------------------------------------------------------------------------------

!   CANCEL PAIRS OF POSITIVE AND NEGATIVE PARTICLES ESCAPING THROUGH THE BOUNDARY
    subroutine boundary_cancel (dt, BC, bref, part, sim)
      implicit none
      integer :: i, j, k, s
      integer, dimension(:), allocatable :: list
      real(8) :: dt, dtf
      type(TBC) :: BC
      type(Tbref) :: bref
      type(Tpart) :: part
      type(Tsim) :: sim

!     perform cancellation for each boundary of type 1 or 2
      do i = 1, sim%Nb
        if ((BC%type(i)==1).OR.(BC%type(i)==2)) then

!         delete all reflecting particles if Np = Nn
          if (bref%N(0,i)==bref%N(1,i)) then
            do j = 1, bref%N(0,i)
              part%del(0,bref%ref(0,i,j)) = .true.
              part%del(1,bref%ref(1,i,j)) = .true.
            end do

!         case: Np > Nn
          else if (bref%N(0,i) < bref%N(1,i)) then

!           create list of positive particles to delete
            allocate (list(bref%N(1,i)))
            list = bref%ref(1, i, 1:bref%N(1,i))
            call shuffle (list, bref%N(1,i))

!           mark particles for deletion
            do j = 1, bref%N(0,i)
              part%del(0, bref%ref(0,i,j)) = .true.
              part%del(1, list(j)) = .true.
            end do

!           diffusively reflect remaining particles
            do j = bref%N(0,i)+1, bref%N(1,i)
              dtf = dt
              call diffuse_reflection (1, list(j), dtf, BC, part, trans, sim)
            end do

!           deallocate list array
            deallocate (list)

!         case: Nn > Np
          else

!           create list of negative particles to delete
            allocate (list(bref%N(0,i)))
            list = bref%ref(0, i, 1:bref%N(0,i))
            call shuffle (list, bref%N(0,i))

!           mark particles for deletion
            do j = 1, bref%N(1,i)
              part%del(0, list(j)) = .true.
              part%del(1, bref%ref(1,i,j)) = .true.
            end do

!           diffusively reflect remaining particles
            do j=bref%N(1,i)+1, bref%N(0,i)
              dtf = dt
              call diffuse_reflection (0, list(j), dtf, BC, part, trans, sim)
            end do

!           deallocate list array
            deallocate (list)
          end if
        end if
      end do

    end subroutine boundary_cancel
!----------------------------------------------------------------------------------------------------------------------------------

!   DRAWS TWO NORMALLY DISTRIBUTED RANDOM VARIABLES USING THE POLAR FORM OF THE BOX MULLER TRANSFORMATION
    subroutine box_Muller (x,y)
      implicit none
      real(8) :: x, y, z

      do
        call random_number(x)
        call random_number(y)
        x=2*x-1
        y=2*y-1
        z=x*x+y*y
        if (z<1.d0) exit
      end do
      z=sqrt((-2*log(z))/z)
      x=sqrt(0.5d0)*x*z
      y=sqrt(0.5d0)*y*z

    end subroutine box_Muller
!----------------------------------------------------------------------------------------------------------------------------------

!   PERFORM DIFFUSE REFLECTION
    recursive subroutine diffuse_reflection (s, i, dt, BC, part, trans, sim)
      implicit none
      logical :: esc
      integer :: s, i, cell, type
      integer, dimension(3,3) :: Itr
      real(8) :: dt, r
      real(8), dimension(2) :: xb
      real(8), dimension(3) :: c
      type(TBC) :: BC
      type(Tpart) :: part
      type(Tprop) :: trans
      type(Tsim) :: sim

!     boundary element to reflect from
      cell = part%cell(s,i)

!     inverse transformation matrix
      Itr = BC%Itr(cell,:,:)

!     determine intersection point and remaining time step
      call intersection (cell, xb, part%x(s,i,:), part%c(s,i,:), dt, sim)

!     sample new velocity and advect to new position

      call sample_fMB_n (c, Itr, BC%u(cell,:), BC%c(cell))
      part%c(s,i,:) = c
      part%x(s,i,1) = xb(1) + c(1)*dt
      part%x(s,i,2) = xb(2) + c(2)*dt

!     determine new cell location
      call locate (esc, cell, part%x(s,i,:), c, sim)
      part%cell(s,i) = cell

!     enforce boundary condition for escaping particles
      if (esc) then

!       determine boundary condition type
        type = BC%type(cell)
        if (type==2) then
          call random_number(r)
          if (r<BC%p(cell)) then
            type = 1
          else
            type = 0
          end if
        end if
        select case (type)

!       specular reflection
        case (0)
          call specular_reflection (s, i, dt, BC, part, trans, sim)

!       diffuse reflection
        case (1)
          call diffuse_reflection (s, i, dt, BC, part, trans, sim)
        end select
      end if

    end subroutine diffuse_reflection
!----------------------------------------------------------------------------------------------------------------------------------

!   EVALUATE MAXWELLIAN
    function eval_fMB (c, fMB)
      implicit none
      real(8) :: eval_fMB
      real(8), dimension(3) :: c
      type(TfMB) :: fMB

      eval_fMB = fMB%n / (srpi*fMB%c)**3 * exp ( - dot_product( c-fMB%u, c-fMB%u ) / fMB%c**2 )

    end function eval_fMB
!----------------------------------------------------------------------------------------------------------------------------------

!   GENERATE ADDITIONAL PARTICLES DUE TO ADVECTION, NONLINEAR VERSION
    subroutine generate_boundary (i, dt, frac, BC, f0, bounds, part, trans, sim, NG, opt)
      implicit none
      logical :: esc, opt, sgn
      integer :: i, N, p, cell, type, s
      integer, dimension(0:1) :: NG
      real(8) :: r, f, H, dt, dtf, a, trial, frac
      real(8), dimension(2) :: x
      real(8), dimension(3) :: V, z, c, b
      type(TBC) :: BC
      type(TfMB) :: f0, fb
      type(Tbounds) :: bounds
      type(Tpart) :: part
      type(Tprop) :: trans
      type(Tsim) :: sim

!     boundary distribution
      fb%n = BC%n(i)
      fb%u = BC%u(i,:)
      fb%c = BC%c(i)

!     apply bound factors
      a    = BC%a(i)   * bounds%Y_a    * bounds%r_Y
      b(1) = BC%b(i,1) * bounds%Y_b(1) * bounds%r_Y
      b(2) = BC%b(i,2) * bounds%Y_b(2) * bounds%r_Y
      b(3) = BC%b(i,3) * bounds%Y_b(2) * bounds%r_Y

!     use number of trial particles as exit condition when opt=.false.
      if (.NOT.opt) then

!       determine the number of trial particles
        trial = dt*frac*BC%Ns(i) * bounds%Y_a * bounds%Y_b(1) * bounds%Y_b(2)**2 * bounds%r_Y**4
        N = int(trial)
        call random_number(r)
        if (r < trial-N) N = N + 1

!       initialize the number of generated particles
        NG = (/0,0/)
      else

!       invert the sign of the desired number of particles to generate
        NG = -NG
      end if

!     generate particles
      p = 0
      do

!       exit criterion
        if (opt) then
          if (sum(NG)==0) exit
        else
          if (p==N) exit
        end if

!       advance do loop counter
        p = p + 1

!       generate random deviates
        call random_number(r)
        H = r*a
        call random_number(r)
        V(1) = r*b(1)
        call random_number(r)
        V(2) = (2.d0*r-1.d0)*b(2)
        call random_number(r)
        V(3) = (2.d0*r-1.d0)*b(3)

!       velocity
        z = V/sqrt(H)
        c = BC%ca(i)*z + BC%u(i,:)/2

!       evaluate distribution
        f = BC%ca(i)**2/BC%na(i) * srpi**3 * c(1) * ( eval_fMB(c,fb) - eval_fMB(c,f0) )

!       particle sign
        s = sign_index(f)

!       acceptance criteria
        sgn = NG(s) /= 0
        if (abs(f)**0.4d0 > H) then
        if ( (.NOT.opt) .OR. sgn ) then

!         transform velocity to global coordinates
          c = matmul(BC%Itr(i,:,:), c)

!         determine random location on the boundary element
          call random_number(r)
          x(1) = BC%x(i,1) + r*BC%dx(i,1)
          call random_number(r)
          x(2) = BC%x(i,2) + r*BC%dx(i,2)
  
!         advect a random fraction of the time step away from the interface
          call random_number(r)
          dtf = r*dt
          x(1) = x(1) + c(1)*dtf
          x(2) = x(2) + c(2)*dtf

!         determine cell number
          call locate (esc, cell, x, c, sim)

!         add particle to simulation
          call add (c, x, s, cell, part)
          NG(s) = NG(s) + 1

!         update bound factors if necessary
          call update_bounds (H, V, a, b, bounds, sum(part%N))

!         enforce boundary condition for escaping particles
          if (esc) then

!           determine boundary condition type
            type = BC%type(cell)
            if (type==2) then
              call random_number(r)
              if (r<BC%p(cell)) then
                type = 1
              else
                type = 0
              end if
            end if
            select case (type)

!           specular reflection
            case (0)
              call specular_reflection (s, part%N(s), dtf, BC, part, trans, sim)

!           diffuse reflection
            case (1)
              call diffuse_reflection (s, part%N(s), dtf, BC, part, trans, sim)
            end select
          end if
        end if
        end if
      end do

    end subroutine generate_boundary
!----------------------------------------------------------------------------------------------------------------------------------

! DETERMINE INTERSECTION POINT OF BOUNDARY AND PARTICLE TRAJECTORY
  subroutine intersection (b, xb, x, c, dt, sim)
    use types
    implicit none
    integer :: b
    real(8) :: dt, cxy
    real(8), dimension(2) :: xb, x
    real(8), dimension(3) :: c
    type(Tsim) :: sim

!     ratio of cx to cy
      cxy = c(1)/c(2)

!     for y = 0 boundary
      if ( b <= sim%Nc(1) ) then
        xb(1) = x(1) - x(2)*cxy
        xb(2) = 0.d0
        dt = x(2)/c(2)

!     for x = Lx boundary
      else if ( b <= sim%Nc(1) + sim%Nc(2) ) then
        xb(1) = sim%L(1)
        xb(2) = x(2) - (x(1)-sim%L(1))/cxy
        dt = (x(1)-sim%L(1))/c(1)

!     for y = Ly boundary
      else if ( b <= 2 * sim%Nc(1) + sim%Nc(2) ) then
        xb(1) = x(1) - (x(2)-sim%L(2))*cxy
        xb(2) = sim%L(2)
        dt = (x(2)-sim%L(2))/c(2)

!     for x = 0 boundary
      else if ( b <= 2 * ( sim%Nc(1) + sim%Nc(2) ) ) then
        xb(1) = 0.d0
        xb(2) = x(2) - x(1)/cxy
        dt = x(1)/c(1)
      end if

    end subroutine intersection
!----------------------------------------------------------------------------------------------------------------------------------

!   LOCATE A PARTICLE TO A PARTICULAR CELL OR BOUNDARY SURFACE
    subroutine locate (esc, cell, x, c, sim)
      use types
      implicit none
      logical :: esc
      integer :: cell
      real(8) :: xb, yb, cxy
      integer, dimension(2) :: n
      real(8), dimension(2) :: x
      real(8), dimension(3) :: c
      type(Tsim) :: sim

!     calculate index position
      n(1) = 1 + floor(sim%Nc(1)*x(1)/sim%L(1))
      n(2) = 1 + floor(sim%Nc(2)*x(2)/sim%L(2))

!     ratio of cx to cy
      cxy = c(1)/c(2)

!     determine boundary cell, if particle escapes
      if (x(1)<0.d0) then

!       x < 0, y < 0
        if (x(2)<0.d0) then
          esc = .true.
          yb = x(2) - x(1)/cxy
          if (yb>0.d0) then
            cell = 2*(sim%Nc(1)+sim%Nc(2)) - floor(sim%Nc(2)*yb/sim%L(2))
          else
            xb = x(1) - x(2)*cxy
            cell = 1 + floor(sim%Nc(1)*xb/sim%L(1))
          end if

!       x < 0, y > Ly
        else if (x(2)>sim%L(2)) then
          esc = .true.
          yb = x(2) - x(1)/cxy
          if (yb<sim%L(2)) then
            cell = 2*(sim%Nc(1)+sim%Nc(2)) - floor(sim%Nc(2)*yb/sim%L(2))
          else
            xb = x(1) - (x(2)-sim%L(2))*cxy
            cell = 2*sim%Nc(1)+sim%Nc(2) - floor(sim%Nc(1)*xb/sim%L(1))
          end if

!       x < 0, 0 < y < Ly
        else
          esc = .true.
          yb = x(2) - x(1)/cxy
          cell = 2*(sim%Nc(1)+sim%Nc(2)) - floor(sim%Nc(2)*yb/sim%L(2))
        end if
      else if (x(1)>sim%L(1)) then

!       x > Lx, y < 0
        if (x(2)<0.d0) then
          esc = .true.
          yb = x(2) - (x(1)-sim%L(1))/cxy
          if (yb>0.d0) then
            cell = sim%Nc(1)+1 + floor(sim%Nc(2)*yb/sim%L(2))
          else
            xb = x(1) - x(2)*cxy
            cell = 1 + floor(sim%Nc(1)*xb/sim%L(1))
          end if

!       x > Lx, y > Ly
        else if (x(2)>sim%L(2)) then
          esc = .true.
          yb = x(2) - (x(1)-sim%L(1))/cxy
          if (yb<sim%L(2)) then
            cell = sim%Nc(1)+1 + floor(sim%Nc(2)*yb/sim%L(2))
          else
            xb = x(1) - (x(2)-sim%L(2))*cxy
            cell = 2*sim%Nc(1)+sim%Nc(2) - floor(sim%Nc(1)*xb/sim%L(1))
          end if

!       x > Lx, 0 < y < Ly
        else
          esc = .true.
          yb = x(2) - (x(1)-sim%L(1))/cxy
          cell = sim%Nc(1)+1 + floor(sim%Nc(2)*yb/sim%L(2))
        end if
      else

!       0 < x < Lx, y < 0
        if (x(2)<0.d0) then
          esc = .true.
          xb = x(1) - x(2)*cxy
          cell = 1 + floor(sim%Nc(1)*xb/sim%L(1))

!       0 < x < Lx, y > Ly
        else if (x(2)>sim%L(2)) then
          esc = .true.
          xb = x(1) - (x(2)-sim%L(2))*cxy
          cell = 2*sim%Nc(1)+sim%Nc(2) - floor(sim%Nc(1)*xb/sim%L(1))

!       0 < x < Lx, 0 < y < Ly
        else
          esc = .false.
          cell = n(1) + sim%Nc(1)*(n(2)-1)
        end if
      end if

    end subroutine locate
!----------------------------------------------------------------------------------------------------------------------------------

!   PRODUCE A RANDOM INTEGER FROM 1 TO N
    function rand_int(N)
      implicit none
      integer :: N, rand_int
      real(8) :: r

      call random_number(r)
      rand_int = int(r*N) + 1

    end function rand_int
!----------------------------------------------------------------------------------------------------------------------------------

!   REMOVE A PARTICLE FROM SIMULATION
    subroutine remove (i, j, part)
      implicit none
      integer :: i, j
      type(Tpart) :: part

!     remove particle
      part%x(i,j,:) = part%x(i,part%N(i),:)
      part%c(i,j,:) = part%c(i,part%N(i),:)
      part%cell(i,j) = part%cell(i,part%N(i))
      part%del(i,j) = .false.
      part%N(i) = part%N(i) - 1

    end subroutine remove
!----------------------------------------------------------------------------------------------------------------------------------

! SAMPLE FROM FLUXAL MAXWELL-BOLTZMANN DISTRIBUTION
  subroutine sample_fMB_n (c, Itr, u, v)
    implicit none
    integer, dimension(3,3) :: Itr
    real(8) :: v, r
    real(8), dimension(3) :: c, u

!   sample from fMB in tangential directions
    call box_Muller (c(2),c(3))

!   sample from fluxal distribution in normal direction
    call random_number(r)
    c(1) = sqrt(-log(r))

!   transform to global coordinates
    c = u + v*matmul(Itr,c)

  end subroutine sample_fMB_n
!----------------------------------------------------------------------------------------------------------------------------------

!   SHUFFLE THE ELEMENTS IN A VECTOR OF LENGTH N
    subroutine shuffle (X, N)
      implicit none
      integer :: N, i, j, Xs
      integer, dimension(:) :: X

      do i=1,N
        j = rand_int(N+1-i) + (i-1)
        Xs = X(j)
        X(j) = X(i)
        X(i) = Xs
      end do

    end subroutine shuffle
!----------------------------------------------------------------------------------------------------------------------------------

!   CHOOSE AN INDEX BASED ON PARTICLE SIGN (1/0=positive/negative)
    function sign_index (x)
      implicit none
      integer :: sign_index
      real(8) :: x

      if (x>=0.d0) then
        sign_index = 1
      else
        sign_index = 0
      end if

    end function sign_index
!----------------------------------------------------------------------------------------------------------------------------------

!   PERFORM SPECULAR REFLECTION
    recursive subroutine specular_reflection (i, j, dt, BC, part, trans, sim)
      implicit none
      logical :: esc
      integer :: i, j, cell, type
      integer, dimension(3) :: n
      real(8) :: dt, r
      real(8), dimension(2) :: xb
      type(TBC) :: BC
      type(Tpart) :: part
      type(Tprop) :: trans
      type(Tsim) :: sim

!     boundary element to reflect from
      cell = part%cell(i,j)

!     inward boundary normal
      n = BC%norm(cell,:)

!     determine intersection point and remaining time step
      call intersection (cell, xb, part%x(i,j,:), part%c(i,j,:), dt, sim)

!     find new velocity and advect to new position
      part%c(i,j,:) = part%c(i,j,:) - 2*dot_product(part%c(i,j,:),n)*n
      part%x(i,j,1) = xb(1) + part%c(i,j,1)*dt
      part%x(i,j,2) = xb(2) + part%c(i,j,2)*dt

!     determine new cell location
      call locate (esc, cell, part%x(i,j,:), part%c(i,j,:), sim)
      part%cell(i,j) = cell

!     enforce boundary condition for escaping particles
      if (esc) then

!       determine boundary condition type
        type = BC%type(cell)
        if (type==2) then
          call random_number(r)
          if (r<BC%p(cell)) then
            type = 1
          else
            type = 0
          end if
        end if
        select case (type)

!       specular reflection
        case (0)
          call specular_reflection (i, j, dt, BC, part, trans, sim)

!       diffuse reflection
        case (1)
          call diffuse_reflection (i, j, dt, BC, part, trans, sim)
        end select
      end if

    end subroutine specular_reflection
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Nmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a48)') '[',tcpu,Np,']: ERROR: Nmax exceeded--increase Nmax parameter'
      write (0,*)
      flush (0)
      stop

    end subroutine stop_exceed_Nmax
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES PER BOUNDARY WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Nbmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a50)') '[',tcpu,Np,']: ERROR: Nbmax exceeded--increase Nbmax parameter'
      write (0,*)
      flush (0)
      stop

    end subroutine stop_exceed_Nbmax
!----------------------------------------------------------------------------------------------------------------------------------

!   UPDATE BOUND FACTORS IF NECESSARY
    subroutine update_bounds (H, V, a, b, bounds, Np)
      implicit none
      logical :: change
      integer :: Np, disp
      real(8) :: H, a, tcpu, Vmax
      real(8), dimension(3) :: V, b
      type(Tbounds) :: bounds

      change = .false.

!     test bound on dependent variable
      if (H > a/bounds%r_Y) then
        bounds%Y_a = (H/a) * bounds%Y_a * bounds%r_Y
        change = .true.
      end if

!     test bound on independent variable, normal component
      if (V(1) > b(1)/bounds%r_Y) then
        bounds%Y_b(1) = (V(1)/b(1)) * bounds%Y_b(1) * bounds%r_Y
        change = .true.
      end if

!     test bound on independent variable, tangential component
      Vmax = max( abs(V(2)/b(2)), abs(V(3)/b(3)) )
      if (Vmax > 1.d0/bounds%r_Y) then
        bounds%Y_b(2) = Vmax * bounds%Y_b(2) * bounds%r_Y
        change = .true.
      end if


!     warn user if bound factors were updated
      if (change) then
        call cpu_time(tcpu)
        write (0, '(a55)') '-------------------------------------------------------'
        write (0, '(a1,es9.3,i9,a36)') '[',tcpu,Np,']: ATTENTION: bound factors changed!'
        write (0, '(a55)') '-------------------------------------------------------'
        flush (0)
      end if

    end subroutine update_bounds
!----------------------------------------------------------------------------------------------------------------------------------

  end subroutine advection


!==================================================================================================================================
! COLLISION ROUTINE
!==================================================================================================================================

! LINEARIZED VHS COLLISION ROUTINE
  subroutine collision (cref, f0, part, prop, sim, time)
    use numerical_constants
    use types
    implicit none
    logical coll, first
    integer :: i, k, s
    type(Tcref) :: cref
    type(TfMB) :: f0
    type(Tpart) :: part
    type(Tprop) :: prop
    type(Tsim) :: sim
    type(Ttime) :: time

!   create cell based cross referencing
    call cross_ref (sim%w, product(sim%Nc), cref, f0, part)

!   perform collision jump events until the collision time step reaches the advection step
    first = .true.
    coll = .false.
    do

!     collision time step
      call step (first, coll, sim%Kn, part, time)
      if (coll) exit
      first = .false.

!     choose random particle
      call choose_P (s, i, cref%M1, part%N, cref%M1max)

!     choose k uniformly from {1, 2, 3, 4}
      k = rand_int(4)

      select case (k)

!     creation event
      case (1:3)
        call creation (s, i, k, cref, f0, part, sim)

!     deletion event
      case (4)
        call deletion (s, i, cref, f0, part, sim)
      end select
    end do

!   correct total sign parity
    call parity (cref, part, sim)

!   correct mass by resampling particles
    call regenerate (cref, f0, part, sim)

  contains
!----------------------------------------------------------------------------------------------------------------------------------

!   ADD PARTICLE TO SIMULATION MAINTAINING CROSS REFERENCING
    subroutine add_ref (c, x, s, w, cell, cref, f0, part)
      implicit none
      integer :: s, cell, ss
      real(8) :: w
      real(8), dimension(2) :: x
      real(8), dimension(3) :: c
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tpart) :: part

!     warn if the maximum number of particles is exceeded
      if (part%N(s) >= Nmax) call stop_exceed_Nmax (sum(part%N(:)))

!     add new velocity sample
      part%N(s) = part%N(s) + 1
      part%c(s,part%N(s),:) = c
      part%x(s,part%N(s),:) = x
      part%cell(s,part%N(s)) = cell
      part%del(s,part%N(s)) = .false.
      cref%G(s,part%N(s)) = .true.

!     warn if the maximum number of particles per cell is exceeded
      if (cref%N(s,cell) >= Ncmax) call stop_exceed_Ncmax (sum(part%N(:)))

!     update cross referencing and M1 values
      cref%N(s,cell) = cref%N(s,cell) + 1
      cref%ref(s,cell,cref%N(s,cell)) = part%N(s)
      cref%idx(s,part%N(s)) = cref%N(s,cell)
      cref%M1(s,part%N(s)) = (1-w)*f0%n + w*eval_M1(c,f0)
      if (cref%M1(s,part%N(s))>cref%M1max) cref%M1max = cref%M1(s,part%N(s))
      cref%M1net = cref%M1net + cref%M1(s,part%N(s))

!     update conservation enforcement tracking
      ss = 2*s-1
      cref%DS = cref%DS + ss
      cref%NG(s) = cref%NG(s) + 1
      cref%Glist(s,1,cref%NG(s)) = part%N(s)
      cref%Gidx(s,part%N(s)) = cref%NG(s)

    end subroutine add_ref
!----------------------------------------------------------------------------------------------------------------------------------

!   DRAWS TWO NORMALLY DISTRIBUTED RANDOM VARIABLES USING THE POLAR FORM OF THE BOX MULLER TRANSFORMATION
    subroutine box_Muller (x,y)
      implicit none
      real(8) :: x, y, z

      do
        call random_number(x)
        call random_number(y)
        x=2.d0*x-1.d0
        y=2.d0*y-1.d0
        z=x*x+y*y
        if (z<1.d0) exit
      end do
      z=sqrt((-2.d0*log(z))/z)
      x=sqrt(0.5d0)*x*z
      y=sqrt(0.5d0)*y*z

    end subroutine box_Muller
!----------------------------------------------------------------------------------------------------------------------------------

!   CHOOSE AN INDEX PROPORTIONAL TO DISCRETE PROBABILITIES FOR SIGNED PARTICLES
    subroutine choose_P (s, i, P, N, Pmax)
      implicit none
      integer :: s, i
      integer, dimension(0:1) :: N
      real(8) :: Pmax, r
      real(8), dimension(:,:), allocatable :: P

!     choose a random index using acceptance rejection
      do

!       choose sign
        s = 0
        call random_number(r)
        if (r < 1.d0*N(1)/(N(0)+N(1))) s = 1

!       choose random particle with appropriate sign
        i = rand_int(N(s))
        call random_number(r)
        if (r < P(s,i)/Pmax) exit
      end do

    end subroutine choose_P
!----------------------------------------------------------------------------------------------------------------------------------

!   CREATE PARTICLE SUBJECT TO ACCEPTANCE CRITERION
    subroutine create (s, i, cp1, cp2, cell, cref, f0, part, sim, opt)
      implicit none
      integer :: cell, i, j, k, s, ss, opt
      real(8) :: w, f0cp, num, den, intf0, r, cjp, cjpt
      real(8), dimension(2) :: x
      real(8), dimension(3) :: cp1, cp2, theta
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tpart) :: part
      type(Tsim) :: sim

!     relative velocity exponent
      w = sim%w

!     evaluate Maxwell-Boltzmann distribution
      f0cp = eval_fMB (cp1, f0)

!     calculate probability of creating particle (Wagner, 3.70)
      num = 0.d0; den = 0.d0
      do ss = 0, 1
      do k = 1, cref%N(ss,cell)

!       index number of current particle according to cross referencing system
        j = cref%ref(ss,cell,k)

!       evaluate the integral of the Maxwellian over the cross plane
        intf0 = int_plane_fMB (part%c(ss,j,:), cp1, f0)

!       auxillary parameter
        theta = proj_plane (cp2, part%c(ss,j,:)-cp1)

!       evaluate velocity magnitudes appearing in (Wagner, 3.70)
        cjp = magnitude (part%c(ss,j,:) - cp1)
        cjpt = magnitude (part%c(ss,j,:) - cp1 + theta)

!       evaluate numerator and denominator for (Wagner, 3.70) seperately
        num = num + (2*ss-1) * (2.d0*intf0/cjp/cjpt**(1.d0-w) - pi*f0cp*cjp**w)
        den = den + 2.d0*intf0/cjp*((1.d0-w)/cjpt + w) + pi*f0cp*(1.d0-w + w*cjp)
      end do
      end do

!     evaluate criterion for creating particle
      call random_number(r)
      if (r < abs(num)/den) then

!       create a new particle according to sign option
        ss = sign_index(num)
        if ((opt==0).OR.(opt==2*ss-1)) then
          x = uniform_x (cref%x(cell,:), sim%dx)
          call add_ref (cp1, x, ss, w, cell, cref, f0, part)
        end if
      end if

    end subroutine create
!----------------------------------------------------------------------------------------------------------------------------------

!   CREATION EVENT FOR COLLISIONS
    subroutine creation (s, i, k, cref, f0, part, sim)
      implicit none
      integer :: i, k, cell, s
      real(8) :: r, crit
      real(8), dimension(3) :: ci, c1, c1p, c2p, thetai
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tpart) :: part
      type(Tsim) :: sim

!     cell number
      cell = part%cell(s,i)

!     particle velocity
      ci = part%c(s,i,:)

!     sample c1 from (Wagner, 3.79)
      call sample_c1 (c1, ci, sim%w, f0)

!     compute c1p depending on k
      if (k==3) then

!       set c1p = c1
        c1p = c1
      else

!       calculate post collision velocity c1p (c2p is not used)
        call post_collision (c1p, c2p, c1, ci)
      end if

!     generate c2p from f0 (c2p + c1p) according to the appropriate criterion
      do
        call sample_fMB (c2p, f0)
        c2p = c2p - c1p

!       iterate only for the cases k = 1, 2
        if (k==3) exit

!       acceptance criterion
        thetai = proj_plane (c2p, ci-c1p)
        call random_number(r)
        crit = ((1.d0 - sim%w)/magnitude(ci - c1p + thetai) + sim%w) / ((1.d0 - sim%w)/magnitude(ci - c1p) + sim%w)
        if (r < crit) exit
      end do

!     create particle subject to acceptance criterion
      call create (s, i, c1p, c2p, cell, cref, f0, part, sim, 0)

    end subroutine creation
!----------------------------------------------------------------------------------------------------------------------------------

!   CREATE CELL BASED CROSS REFERENCING
    subroutine cross_ref (w, Ncell, cref, f0, part)
      implicit none
      integer :: Ncell, i, j, k, s
      real(8) :: w
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tpart) :: part

!     initialize cross referencing
      do k = 0, 1
      do i = 1, Ncell
        cref%N(k,i) = 0
      end do
      end do

!     initialize net and max M1
      cref%M1net = 0.d0
      cref%M1max = 0.d0

!     create cross reference list
      do k = 0, 1
      s = 2*k-1
      do j = 1, part%N(k)
        i = part%cell(k,j)

!       warn if the maximum number of particles per cell is exceeded
        if (cref%N(k,i) >= Ncmax) call stop_exceed_Ncmax (sum(part%N(:)))

        cref%N(k,i) = cref%N(k,i) + 1
        cref%ref(k,i,cref%N(k,i)) = j
        cref%idx(k,j) = cref%N(k,i)

!       compute M1 values
        cref%M1(k,j) = (1-w)*f0%n + w*eval_M1(part%c(k,j,:),f0)
        if (cref%M1(k,j)>cref%M1max) cref%M1max = cref%M1(k,j)
        cref%M1net = cref%M1net + cref%M1(k,j)
      end do
      end do

!     initialize conservation cross-referencing
      cref%NG(:) = (/0,0/)
      do k = 0, 1
      do j = 1, part%N(k)
        cref%G(k,j) = .false.
      end do
      end do

    end subroutine cross_ref
!----------------------------------------------------------------------------------------------------------------------------------

!   PARTICLE DELETION EVENT
    subroutine deletion (s, i, cref, f0, part, sim)
      implicit none
      integer :: i, cell, s
      real(8) :: r, crit, ci1, w
      real(8), dimension(3) :: ci, c1
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tpart) :: part
      type(Tsim) :: sim

!     velocity dependence of collision section
      w = sim%w

!     cell number
      cell = part%cell(s,i)

!     particle velocity
      ci = part%c(s,i,:)

!     sample c1 from (Wagner, 3.79)
      call sample_c1 (c1, ci, w, f0)

!     evaluate deletion probability
      ci1 = magnitude(ci-c1)
      crit = ci1**w / (1.d0-w + w*ci1)

!     delete particle subject to deletion criterion
      call random_number(r)
      if (r < crit) call remove_ref (s, i, cref, part)

    end subroutine deletion
!----------------------------------------------------------------------------------------------------------------------------------

!   DELETE RANDOM PARTICLE FROM LIST OF PREVIOUSLY-GENERATED PARTICLES
    subroutine delete (cref, part, sim, opt)
      implicit none
      integer :: s, i, opt
      real(8) :: r
      type(Tcref) :: cref
      type(Tpart) :: part
      type(Tsim) :: sim

!     choose sign at random, if not specified
      if (opt==0) then
        s = 0
        call random_number(r)
        if (r < 1.d0*cref%NG(1)/(cref%NG(0)+cref%NG(1))) s = 1

!     otherwise, assign to specified value
      else
        s = (opt+1)/2
      end if

!     choose index
      i = rand_int(cref%NG(s))
      i = cref%Glist(s,1,i)

!     remove particle
      call remove_ref (s, i, cref, part)

    end subroutine delete
!----------------------------------------------------------------------------------------------------------------------------------

!   EVALUATE MAXWELLIAN
    function eval_fMB (c, f0)
      implicit none
      real(8) :: eval_fMB
      real(8), dimension(3) :: c
      type(TfMB) :: f0

      eval_fMB = f0%n/(srpi*f0%c)**3 * exp (-dot_product(c-f0%u,c-f0%u)/f0%c**2)

    end function eval_fMB
!----------------------------------------------------------------------------------------------------------------------------------

!   NONCENTRAL MOMENT OF A MAXWELLIAN: int{ ||cp - c|| f0(c) }
    function eval_M1 (c, f0)
      implicit none
      real(8) :: eval_M1, U, U2
      real(8), dimension(3) :: c
      type(TfMB) :: f0

      U2 = dot_product(c-f0%u,c-f0%u)/f0%c**2
      U = sqrt(U2)

      eval_M1 = f0%n*f0%c*(exp(-U2)/srpi + (0.5d0/U + U)*erf(U))

    end function eval_M1
!----------------------------------------------------------------------------------------------------------------------------------

!   PERFORM PARTICLE GENERATION STEPS UNTIL ONE ADDITIONAL PARTICLE IS GENERATED
    subroutine generate (cref, f0, part, sim, opt)
      implicit none
      integer :: i, k, cell, s, Np0, opt
      real(8) :: r, crit
      real(8), dimension(3) :: ci, c1, c1p, c2p, thetai
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tpart) :: part
      type(Tsim) :: sim

!     cycle until a single particle is generated
      Np0 = sum(part%N)
      do

!       exit condition
        if ((sum(part%N)-Np0)==1) exit

!       choose random particle
        call choose_P (s, i, cref%M1, part%N, cref%M1max)

!       choose k uniformly from {1, 2, 3}
        k = rand_int(3)

!       cell number
        cell = part%cell(s,i)

!       particle velocity
        ci = part%c(s,i,:)

!       sample c1 from (Wagner, 3.79)
        call sample_c1 (c1, ci, sim%w, f0)

!       compute c1p depending on k
        if (k==3) then

!         set c1p = c1
          c1p = c1
        else

!         calculate post collision velocity c1p (c2p is not used)
          call post_collision (c1p, c2p, c1, ci)
        end if

!       generate c2p from f0 (c2p + c1p) according to the appropriate criterion
        do
          call sample_fMB (c2p, f0)
          c2p = c2p - c1p

!         iterate only for the cases k = 1, 2
          if (k==3) exit

!         acceptance criterion
          thetai = proj_plane (c2p, ci-c1p)
          call random_number(r)
          crit = ((1.d0 - sim%w)/magnitude(ci - c1p + thetai) + sim%w) / ((1.d0 - sim%w)/magnitude(ci - c1p) + sim%w)
          if (r < crit) exit
        end do

!       create particle subject to acceptance criterion
        call create (s, i, c1p, c2p, cell, cref, f0, part, sim, opt)
      end do

    end subroutine generate
!----------------------------------------------------------------------------------------------------------------------------------

!   INTEGRATE MAXWELLIAN OVER CROSS PLANE
    function int_plane_fMB (c, cp, f0)
      implicit none
      real(8) :: int_plane_fMB
      real(8), dimension(3) :: c, cp
      type(TfMB) :: f0

      int_plane_fMB = f0%n/srpi/f0%c * exp( -dot_product(f0%u-cp,c-cp)**2/f0%c**2/dot_product(c-cp,c-cp) )

    end function int_plane_fMB
!----------------------------------------------------------------------------------------------------------------------------------

!   VECTOR MAGNINTUDE
    function magnitude (c)
      implicit none
      real(8) :: magnitude
      real(8), dimension(3) :: c

      magnitude = sqrt(dot_product(c,c))

    end function magnitude
!----------------------------------------------------------------------------------------------------------------------------------

!   CORRECT TOTAL SIGN PARITY
    subroutine parity (cref, part, sim)
      implicit none
      real(8) :: r
      type(Tcref) :: cref
      type(Tpart) :: part
      type(Tsim) :: sim

!     for odd parity, change it to even (if possible)
      if ((mod(abs(cref%DS),2)==1).AND.(sum(cref%NG(:))>0)) then

!       with equal probability increase/decrease generated particle population
        call random_number(r)
        if (r<0.5d0) then

!         delete previously generated particle
          call delete (cref, part, sim, 0)
        else

!         otherwise, generate additional particle
          call generate (cref, f0, part, sim, 0)
        end if
      end if

    end subroutine parity
!----------------------------------------------------------------------------------------------------------------------------------

!   CALCULATE POST COLLISION VELOCITIES
    subroutine post_collision (c1p, c2p, c1, c2)
      implicit none
      real(8) :: cr
      real(8), dimension(3) :: c1p, c2p, c1, c2, e, cavg

      cavg = 0.5d0*(c1 + c2)
      cr = magnitude(c1-c2)
      call sample_sphere (e)

      c1p = cavg + 0.5d0*cr*e
      c2p = cavg - 0.5d0*cr*e

    end subroutine post_collision
!----------------------------------------------------------------------------------------------------------------------------------

!   PROJECT VECTOR x ONTO THE PLANE THROUGH THE ORIGIN PERPENDICULAR TO v
    function proj_plane (x, v)
      implicit none
      real(8), dimension(3) :: x, v, proj_plane

      proj_plane = x - dot_product(x,v)/dot_product(v,v)*v

    end function proj_plane
!----------------------------------------------------------------------------------------------------------------------------------

!   PRODUCE A RANDOM INTEGER FROM 1 TO N
    function rand_int(N)
      implicit none
      integer :: N, rand_int
      real(8) :: r

      call random_number(r)
      rand_int = int(r*N) + 1

    end function rand_int
!----------------------------------------------------------------------------------------------------------------------------------

!   PERFORM REGENERATE PREVIOUSLY GENERATED PARTICLES WITH OPPOSITE SIGN UNTIL OPTIMAL SIGN DEFICIT IS OBTAINED
    subroutine regenerate (cref, f0, part, sim)
      implicit none
      integer :: s, DS
      real(8) :: r
      type(Tcref) :: cref
      type(TfMB) :: f0
      type(Tpart) :: part
      type(Tsim) :: sim

!     perform this step only if the mass deficit is nonzero and there are generated particles to resample
      if ((cref%DS/=0).AND.(sum(cref%NG(:))>0)) then

!       determine optimal sign deficit
        s = sign(1,cref%DS)
        if (s==-1) then
          DS = min(0,cref%DS+2*cref%NG(0))
        else
          DS = max(0,cref%DS-2*cref%NG(1))
        end if

!       perform regeneration step until optimal sign deficit is obtained
        do

!         check sign deficit, and exit (if appropriate)
          if (cref%DS==DS) exit

!         delete/generate particle in random order
          call random_number(r)
          if (r<0.5D0) then

!           delete first, then generate
            call delete (cref, part, sim, s)
            call generate (cref, f0, part, sim, -s)
          else

!           generate first, then delete
            call generate (cref, f0, part, sim, -s)
            call delete (cref, part, sim, s)
          end if
        end do
      end if

    end subroutine regenerate
!----------------------------------------------------------------------------------------------------------------------------------

!   REMOVE A PARTICLE FROM SIMULATION MAINTAINING CROSS REFERENCING
    subroutine remove_ref (s, i, cref, part)
      implicit none
      integer :: i, j, k, s, ss
      real(8), dimension(3) :: c
      type(Tcref) :: cref
      type(Tpart) :: part

!     particle sign and velocity
      ss = 2*s-1
      c = part%c(s,i,:)

!     update cross referencing and M1 for deleted particle
      j = part%cell(s,i)
      k = cref%ref(s,j,cref%N(s,j))
      cref%ref(s,j,cref%idx(s,i)) = k
      if (i/=part%N(s)) cref%ref(s,part%cell(s,part%N(s)),cref%idx(s,part%N(s))) = i
      cref%idx(s,k) = cref%idx(s,i)
      if (i/=part%N(s)) cref%idx(s,i) = cref%idx(s,part%N(s))
      cref%N(s,j) = cref%N(s,j) - 1
      cref%M1net = cref%M1net - cref%M1(s,i)

!     reassign highest numbered particle in the simulation
      if (i /= part%N(s)) then
        part%x(s,i,:) = part%x(s,part%N(s),:)
        part%c(s,i,:) = part%c(s,part%N(s),:)
        part%cell(s,i) = part%cell(s,part%N(s))
        cref%idx(s,i) = cref%idx(s,part%N(s))
        cref%ref(s,part%cell(s,i), cref%idx(s,i)) = i
        cref%M1(s,i) = cref%M1(s,part%N(s))
      end if

!     update conservation enforcement tracking
      cref%DS = cref%DS - ss

!     correct list of generated particles if affected
      if ( cref%G(s,i) .OR. cref%G(s,part%N(s)) ) then

!       if both deleted and highest-numbered particle were in list of generated particles
        if ( cref%G(s,i) .AND. cref%G(s,part%N(s)) ) then
          if ( part%N(s) /= cref%Glist(s,1,cref%NG(s)) ) then
            cref%Glist(s,1,cref%Gidx(s,part%N(s))) = i
            cref%Glist(s,1,cref%Gidx(s,i)) = cref%Glist(s,1,cref%NG(s))
            cref%Gidx(s,cref%Glist(s,1,cref%NG(s))) = cref%Gidx(s,i)
            cref%Gidx(s,i) = cref%Gidx(s,part%N(s))
          end if
          cref%NG(s) = cref%NG(s) - 1

!       if only the deleted particle was in list of generated particles
        else if ( cref%G(s,i) ) then
          cref%Glist(s,1,cref%Gidx(s,i)) = cref%Glist(s,1,cref%NG(s))
          cref%Gidx(s,cref%Glist(s,1,cref%NG(s))) = cref%Gidx(s,i)
          cref%NG(s) = cref%NG(s) - 1
          cref%G(s,i) = .false.

!       if only the highest-numbered particle was in the list of generated particles
        else
          cref%Glist(s,1,cref%Gidx(s,part%N(s))) = i
          cref%Gidx(s,i) = cref%Gidx(s,part%N(s))
          cref%G(s,i) = .true.
        end if
      end if

!     decreased number of particles in the simulation
      part%N(s) = part%N(s) - 1

    end subroutine remove_ref
!----------------------------------------------------------------------------------------------------------------------------------

!   SAMPLE FROM DISTRIBUTION (Wagner, 3.79)
    subroutine sample_c1 (c1, ci, w, f0)
      implicit none
      real(8) :: w, P, r
      real(8), dimension(3) :: ci, c1
      type(TfMB) :: f0

!     determine which distribution to sample from
      P = (1.d0-sim%w)/(1.d0-sim%w + sim%w*eval_M1(ci,f0))
      call random_number(r)
      if (r<P) then

!       sample from f0
        call sample_fMB (c1, f0)
      else

!       sample from ||ci - c||f0
        call sample_M1 (c1, ci, f0)
      end if

    end subroutine sample_c1
!----------------------------------------------------------------------------------------------------------------------------------

!   PRODUCE A SINGLE SAMPLE FROM A GAUSSIAN DISTRIBUTION
    subroutine sample_fMB (c, f0)
      implicit none
      real(8) :: c3
      real(8), dimension(3) :: c
      type(TfMB) :: f0

!     produce sample
      call box_Muller (c(1),c(2))
      call box_muller (c(3),c3)

!     scale mean and variance
      c = f0%u + f0%c*c

    end subroutine sample_fMB
!----------------------------------------------------------------------------------------------------------------------------------

!   SAMPLE FROM THE DISTRIBUTION ||ci - c||f0
    subroutine sample_M1 (c1, ci, f0)
      implicit none
      real(8) :: P, r, ci0, c1i, c10
      real(8), dimension(3) :: c1, ci
      type(TfMB) :: f0

!     repeat until acceptance requirement is met
      do

!       probability of generating a trial sample from ||c - u0||f0
        ci0 = magnitude(ci-f0%u)
        P = 2.d0/(2.d0 + srpi*ci0/f0%c)

!       determine trial sample distribution
        call random_number(r)
        if (r<P) then

!         sample from ||c - u0||f0
          call sample_M10 (c1, f0)
        else

!         sample from f0
          call sample_fMB (c1, f0)
        end if

!       acceptance criterion
        call random_number(r)
        c1i = magnitude(c1-ci)
        c10 = magnitude(c1-f0%u)
        if (r < c1i/(c10+ci0)) exit
      end do

    end subroutine sample_M1
!----------------------------------------------------------------------------------------------------------------------------------

!   SAMPLE FROM THE DISTRIBUTION ||c - u0||f0
    subroutine sample_M10 (c1, f0)
      implicit none
      real(8) :: H, x, r
      real(8), dimension(3) :: c1, e
      type(TfMB) :: f0

!     sample from distribution x^3 e^(-x^2)
      do
        call random_number(r)
        H = 0.551819161757163d0 * r
        call random_number(r)
        x = 1.103638323514327d0 * r/sqrt(H)

!       ratio-of-uniforms acceptance criterion
        if ( H < (x**3*exp(-x**2))**0.666666666666666d0 ) exit
      end do

!     random vector direction
      call sample_sphere (e)

!     shift and rescale
      c1 = f0%c*x*e + f0%u

    end subroutine sample_M10
!----------------------------------------------------------------------------------------------------------------------------------

!   RANDOMLY SAMPLE THE UNIFORM SPHERE
    subroutine sample_sphere (e)
      implicit none
      real(8) :: cos_chi, sin_chi, azimuth, r
      real(8), dimension(3) :: e

      call random_number (r)
      cos_chi = 2.d0*r - 1.d0
      sin_chi = sqrt(1.d0-cos_chi**2)
      call random_number (r)
      azimuth = 2.d0*pi*r
      e(1) = sin_chi*cos(azimuth)
      e(2) = sin_chi*sin(azimuth)
      e(3) = cos_chi

    end subroutine sample_sphere
!----------------------------------------------------------------------------------------------------------------------------------

!   SHUFFLE THE ELEMENTS IN A VECTOR OF LENGTH N
    subroutine shuffle (X, N)
      implicit none
      integer :: N, i, j, Xs
      integer, dimension(:) :: X

      do i=1,N
        j = rand_int(N+1-i) + (i-1)
        Xs = X(j)
        X(j) = X(i)
        X(i) = Xs
      end do

    end subroutine shuffle
!----------------------------------------------------------------------------------------------------------------------------------

!   CHOOSE AN INDEX BASED ON PARTICLE SIGN (1/0=positive/negative)
    function sign_index (x)
      implicit none
      integer :: sign_index
      real(8) :: x

      if (x>=0.d0) then
        sign_index = 1
      else
        sign_index = 0
      end if

    end function sign_index
!----------------------------------------------------------------------------------------------------------------------------------

!   ADD EXPONENTIALLY DISTRIBUTED TIME STEP TO THE CURRENT COLLISION TIME
    subroutine step (first, coll, Kn, part, time)
      implicit none
      logical :: coll, first
      real(8) :: Kn, r, a
      type(Tpart) :: part
      type(Ttime) :: time

!     check to see if the simulation is empty
      if (cref%M1net < eps1) then

!       advance cell a full time step if it is empty
        time%tc = time%t
        coll = .true.

!     normal routine
      else

!       create new time step only if one is not available from previous advection step
        if ( (.NOT.first) .OR. (time%dtc < 0.d0) ) then

!         time step parameter
          a = 16.d0*pi*Kn*cref%M1net

!         generate time step
          call random_number(r)
          time%dtc = -log(r)/a
        end if

!       advance collision time only if it does not exceed advection time
        if (time%tc + time%dtc < time%t) then
          time%tc = time%tc + time%dtc
        else
          coll = .true.
        end if
      end if

    end subroutine step
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Nmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a48)') '[',tcpu,Np,']: ERROR: Nmax exceeded--increase Nmax parameter'
      write (0, '(a55)') '-------------------------------------------------------'
      flush (0)
      stop

    end subroutine stop_exceed_Nmax
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES PER CELL WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Ncmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a50)') '[',tcpu,Np,']: ERROR: Ncmax exceeded--increase Ncmax parameter'
      write (0, '(a55)') '-------------------------------------------------------'
      flush (0)
      stop

    end subroutine stop_exceed_Ncmax
!----------------------------------------------------------------------------------------------------------------------------------

!   GENERATE UNIFORMLY DISTRIBUTED SPATIAL POSITION WITHIN A CELL
    function uniform_x (x0, dx)
      implicit none
      real(8) :: r
      real(8), dimension(2) :: x0, dx, uniform_x

      call random_number(r)
      uniform_x(1) = x0(1) + r*dx(1)
      call random_number(r)
      uniform_x(2) = x0(2) + r*dx(2)

    end function uniform_x
!----------------------------------------------------------------------------------------------------------------------------------

  end subroutine collision


!==================================================================================================================================
! EFFECTIVE BODY FORCE ROUTINE
!==================================================================================================================================

! GENERATE PARTICLES TO ACCOMODATE VARIOUS KINDS OF EFFECTIVE BODY FORCE TERMS
  subroutine force (f0, part, sim)
    use types
    use numerical_constants
    implicit none
    integer :: j
    integer, dimension(0:1) :: NG
    type(TfMB) :: f0
    type(Tpart) :: part
    type(Tsim) :: sim

!   particle generation routine with mass conservation
    call generate_force (0.5d0, f0, part, sim, NG, .false.)
    j = NG(0)
    NG(0) = NG(1)
    NG(1) = j
    call generate_force (0.5d0, f0, part, sim, NG, .true.)

  contains
!----------------------------------------------------------------------------------------------------------------------------------

!   ADD PARTICLE TO SIMULATION
    subroutine add (c, x, s, cell, part)
      implicit none
      integer :: s, cell
      real(8), dimension(2) :: x
      real(8), dimension(3) :: c
      type(Tpart) :: part

!     warn if the maximum number of particles is exceeded
      if (maxval(part%N)>=Nmax) call stop_exceed_Nmax (sum(part%N))

!     add new velocity sample
      part%N(s) = part%N(s) + 1
      part%c(s, part%N(s),:) = c
      part%x(s, part%N(s),:) = x
      part%cell(s, part%N(s)) = cell
      part%del(s, part%N(s)) = .false.

    end subroutine add
!----------------------------------------------------------------------------------------------------------------------------------

!   GENERATE PARTICLES FOR EFFECTIVE BODY FORCE TERMS
    subroutine generate_force (frac, f0, part, sim, NG, opt)
      implicit none
      logical :: esc, opt, sgn
      integer :: N, p, s, cell
      integer, dimension(0:1) :: NG
      real(8) :: r, H, f, c2, frac
      real(8), dimension(2) :: x
      real(8), dimension(3) :: V, c
      type(TfMB) :: f0
      type(Tpart) :: part
      type(Tsim) :: sim

!     use number of trial particles as exit condition when opt=.false.
      if (.NOT.opt) then

!       determine the number of trial particles
        N = int(sim%Nf*frac)
        call random_number(r)
        if (r < sim%Nf*frac-N) N = N + 1

!       initialize the number of generated particles
        NG = (/0,0/)
      else

!       invert the sign of the desired number of particles to generate
        NG = -NG
      end if

!     generate particles
      p = 0
      do

!       exit criterion
        if (opt) then
          if (sum(NG)==0) exit
        else
          if (p==N) exit
        end if

!       advance do loop counter
        p = p + 1

!       generate random deviates
        call random_number(r)
        H = r*sim%af
        call random_number(r)
        V(1) = (2*r-1)*sim%bf(1)
        call random_number(r)
        V(2) = (2*r-1)*sim%bf(2)
        call random_number(r)
        V(3) = (2*r-1)*sim%bf(3)

!       velocity
        c = V/sqrt(H)

!       evaluate distribution
        c2 = dot_product(c,c)
        f =  ( sim%Pz*c(3) + sim%Tz*c(3)*(2.5d0-c2) ) * exp(-c2)

!       particle sign
        s = sign_index(f)

!       acceptance criteria
        sgn = NG(s) /= 0
        if (abs(f)**0.4 > H) then
        if ( (.NOT.opt) .OR. sgn ) then

!         adjust particle velocity based on reference distribution
          c = f0%c*c + f0%u

!         particle position
          call random_number(r)
          x(1) = r*sim%L(1)
          call random_number(r)
          x(2) = r*sim%L(2)

!         determine cell number
          call locate (esc, cell, x, c, sim)

!         add particle to simulation
          call add (c, x, s, cell, part)
          NG(s) = NG(s) + 1
        end if
        end if
      end do

    end subroutine generate_force
!----------------------------------------------------------------------------------------------------------------------------------

!   LOCATE A PARTICLE TO A PARTICULAR CELL OR BOUNDARY SURFACE
    subroutine locate (esc, cell, x, c, sim)
      use types
      implicit none
      logical :: esc
      integer :: cell
      real(8) :: xb, yb, cxy
      integer, dimension(2) :: n
      real(8), dimension(2) :: x
      real(8), dimension(3) :: c
      type(Tsim) :: sim

!     calculate index position
      n(1) = 1 + floor(sim%Nc(1)*x(1)/sim%L(1))
      n(2) = 1 + floor(sim%Nc(2)*x(2)/sim%L(2))

!     ratio of cx to cy
      cxy = c(1)/c(2)

!     determine boundary cell, if particle escapes
      if (x(1)<0.d0) then

!       x < 0, y < 0
        if (x(2)<0.d0) then
          esc = .true.
          yb = x(2) - x(1)/cxy
          if (yb>0.d0) then
            cell = 2*(sim%Nc(1)+sim%Nc(2)) - floor(sim%Nc(2)*yb/sim%L(2))
          else
            xb = x(1) - x(2)*cxy
            cell = 1 + floor(sim%Nc(1)*xb/sim%L(1))
          end if

!       x < 0, y > Ly
        else if (x(2)>sim%L(2)) then
          esc = .true.
          yb = x(2) - x(1)/cxy
          if (yb<sim%L(2)) then
            cell = 2*(sim%Nc(1)+sim%Nc(2)) - floor(sim%Nc(2)*yb/sim%L(2))
          else
            xb = x(1) - (x(2)-sim%L(2))*cxy
            cell = 2*sim%Nc(1)+sim%Nc(2) - floor(sim%Nc(1)*xb/sim%L(1))
          end if

!       x < 0, 0 < y < Ly
        else
          esc = .true.
          yb = x(2) - x(1)/cxy
          cell = 2*(sim%Nc(1)+sim%Nc(2)) - floor(sim%Nc(2)*yb/sim%L(2))
        end if
      else if (x(1)>sim%L(1)) then

!       x > Lx, y < 0
        if (x(2)<0.d0) then
          esc = .true.
          yb = x(2) - (x(1)-sim%L(1))/cxy
          if (yb>0.d0) then
            cell = sim%Nc(1)+1 + floor(sim%Nc(2)*yb/sim%L(2))
          else
            xb = x(1) - x(2)*cxy
            cell = 1 + floor(sim%Nc(1)*xb/sim%L(1))
          end if

!       x > Lx, y > Ly
        else if (x(2)>sim%L(2)) then
          esc = .true.
          yb = x(2) - (x(1)-sim%L(1))/cxy
          if (yb<sim%L(2)) then
            cell = sim%Nc(1)+1 + floor(sim%Nc(2)*yb/sim%L(2))
          else
            xb = x(1) - (x(2)-sim%L(2))*cxy
            cell = 2*sim%Nc(1)+sim%Nc(2) - floor(sim%Nc(1)*xb/sim%L(1))
          end if

!       x > Lx, 0 < y < Ly
        else
          esc = .true.
          yb = x(2) - (x(1)-sim%L(1))/cxy
          cell = sim%Nc(1)+1 + floor(sim%Nc(2)*yb/sim%L(2))
        end if
      else

!       0 < x < Lx, y < 0
        if (x(2)<0.d0) then
          esc = .true.
          xb = x(1) - x(2)*cxy
          cell = 1 + floor(sim%Nc(1)*xb/sim%L(1))

!       0 < x < Lx, y > Ly
        else if (x(2)>sim%L(2)) then
          esc = .true.
          xb = x(1) - (x(2)-sim%L(2))*cxy
          cell = 2*sim%Nc(1)+sim%Nc(2) - floor(sim%Nc(1)*xb/sim%L(1))

!       0 < x < Lx, 0 < y < Ly
        else
          esc = .false.
          cell = n(1) + sim%Nc(1)*(n(2)-1)
        end if
      end if

    end subroutine locate
!----------------------------------------------------------------------------------------------------------------------------------

!   CHOOSE AN INDEX BASED ON PARTICLE SIGN (1/0=positive/negative)
    function sign_index (x)
      implicit none
      integer :: sign_index
      real(8) :: x

      if (x>=0.d0) then
        sign_index = 1
      else
        sign_index = 0
      end if

    end function sign_index
!----------------------------------------------------------------------------------------------------------------------------------

!   WARN USER THAT THE MAXIMUM NUMBER OF PARTICLES WAS EXCEEEDED AND STOP EXECUTION
    subroutine stop_exceed_Nmax (Np)
      implicit none
      integer :: Np
      real(8) :: tcpu

      call cpu_time(tcpu)
      write (0, '(a55)') '-------------------------------------------------------'
      write (0, '(a1,es9.3,i9,a48)') '[',tcpu,Np,']: ERROR: Nmax exceeded--increase Nmax parameter'
      write (0, '(a55)') '-------------------------------------------------------'
      flush (0)
      stop

    end subroutine stop_exceed_Nmax
!----------------------------------------------------------------------------------------------------------------------------------

  end subroutine force

!==================================================================================================================================
